/*
 * arch/arm/mach-tcc92x/cpufreq.c
 *
 * TCC92xx cpufreq driver
 *
 * Copyright (C) 2009 Telechips, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/suspend.h>
#include <linux/regulator/machine.h>
#include <linux/regulator/consumer.h>
#include <bsp.h>

#define TCC_TRANSITION_LATENCY		100

#if defined(CONFIG_DRAM_DDR2)
#define TCC_CPU_FREQ_ONDEMAND_MAX	800000
#endif

#define CALLED_BY_CPUFREQ_C_FILE
#if defined(CONFIG_DRAM_MDDR)
#include "tcc_clocktbl_lpddr.h"
#elif defined(CONFIG_DRAM_DDR2)
#include "tcc_clocktbl_ddr2.h"
#elif defined(CONFIG_DRAM_DDR3)
#include "tcc_clocktbl_ddr3.h"
#else
#error Select the DDR type
#endif

struct tcc_pll_table_t {
	unsigned int cpu_freq;
	unsigned int pll_freq;
};

struct tcc_voltage_table_t {
	unsigned int cpu_freq;
	unsigned int voltage;
};

static struct tcc_pll_table_t tcc_pll_table[] = {
	{ 144000, 144000 },
	{ 200000, 204000 },
	{ 300000, 304000 },
	{ 400000, 400000 },
	{ 500000, 500000 },
	{ 600000, 600000 },
	{ 700000, 700000 },
	{ 800000, 800000 },
};

static struct tcc_voltage_table_t tcc_voltage_table[] = {
	{  36000, 1050000 },
	{ 294000, 1100000 },
	{ 425250, 1150000 },
	{ 486000, 1200000 },
	{ 500000, 1200000 },
	{ 540000, 1250000 },
	{ 600000, 1300000 },
	{ 660000, 1350000 },
	{ 704000, 1400000 },
	{ 720000, 1400000 },
};

static struct tcc_freq_table_t tcc_freq_old_table = {
	       0,       0,      0,       0,       0,     0,     0,       0
};

extern struct tcc_freq_table_t gtClockLimitTable[];

struct tcc_freq_limit_table_t {
	struct tcc_freq_table_t freq;
	int usecount;
};

static struct tcc_freq_limit_table_t tcc_freq_limit_table[TCC_FREQ_LIMIT_MAX] = {{{0,0,0,0,0,0,0,0}, 0},};
static struct tcc_freq_table_t tcc_freq_curr_limit_table;
static unsigned int tcc_limitclocktbl_flag = 0;

#define NUM_VOLTAGES	ARRAY_SIZE(tcc_voltage_table)
#define NUM_PLLS		ARRAY_SIZE(tcc_pll_table)
#define NUM_FREQS		ARRAY_SIZE(gtClockLimitTable)

static struct cpufreq_frequency_table tcc_cpufreq_table[NUM_FREQS + 1];

static struct clk *pll0_clk;
static struct clk *cpu_clk;
static struct clk *mem_clk;
static struct clk *io_clk;
static struct clk *ddi_clk;
static struct clk *gpu_clk;
static struct clk *smu_clk;
static struct clk *vcod_clk;
static struct clk *vbus_clk;
#if defined(CONFIG_USE_ISP)
static struct clk *isp_bus_clk;
#else
static struct clk *cam_bus_clk;
#endif

#ifdef CONFIG_REGULATOR
static struct regulator *vdd_core;
#endif
static int curr_voltage = 0;

/**
 * Get current CPU frequency in kHz
 */
static unsigned int tcc_cpufreq_get(unsigned int cpu)
{
	return clk_get_rate(cpu_clk) / 1000;
}

static int tcc_cpufreq_set_voltage(int uV)
{
	int ret = 0;

#if defined(CONFIG_REGULATOR)
	if (!vdd_core)
		return -EINVAL;

	ret = regulator_set_voltage(vdd_core, uV, uV);
	if (ret == 0)
		curr_voltage = uV;
#endif
	return ret;
}

int tcc_cpufreq_get_voltage(void)
{
	int ret = 0;

#if defined(CONFIG_REGULATOR)
	ret = regulator_get_voltage(vdd_core);
#endif

	return ret;
}

void tcc_cpufreq_set_voltage_limit(int min_uV, int type)
{
	printk("######### Not used function: %s #########\n", __func__);
}
EXPORT_SYMBOL(tcc_cpufreq_set_voltage_limit);

void tcc_cpufreq_set_min_freq(unsigned int freq)
{
	printk("######### Not used function: %s #########\n", __func__);
}
EXPORT_SYMBOL(tcc_cpufreq_set_min_freq);

static unsigned int tcc_cpufreq_get_pll_table(unsigned int cpu_clk)
{
	int i;

	for (i=0 ; i<NUM_PLLS ; i++) {
		if (tcc_pll_table[i].cpu_freq >= cpu_clk) {
			i++;
			break;
		}
	}

	return tcc_pll_table[i-1].pll_freq;
}

static int tcc_cpufreq_get_voltage_by_cpufreq(unsigned int cpu_clk)
{
	int i;

	for (i=(NUM_VOLTAGES-1) ; i>0 ; i--) {
		if (tcc_voltage_table[i].cpu_freq <= cpu_clk)
			break;
	}

	return tcc_voltage_table[i].voltage;
}

static int tcc_cpufreq_get_voltage_table(struct tcc_freq_table_t *clk_tbl)
{
	int i;
	
	for (i=0 ; i<NUM_FREQS ; i++) {
		if (gtClockLimitTable[i].cpu_freq >= clk_tbl->cpu_freq &&
			gtClockLimitTable[i].ddi_freq >= clk_tbl->ddi_freq &&
			gtClockLimitTable[i].mem_freq >= clk_tbl->mem_freq &&
			gtClockLimitTable[i].gpu_freq >= clk_tbl->gpu_freq &&
			gtClockLimitTable[i].io_freq >= clk_tbl->io_freq)
			break;
	}

	if (i >= NUM_FREQS)
		i = NUM_FREQS-1;

	return tcc_cpufreq_get_voltage_by_cpufreq(gtClockLimitTable[i].cpu_freq);
}

extern struct mutex mutex_tccnfc;
static int tcc_cpufreq_set_clock_table(struct tcc_freq_table_t *curr_clk_tbl)
{
	int new_voltage;
	int ret = 0;

	new_voltage = tcc_cpufreq_get_voltage_table(curr_clk_tbl);

//	printk("cpufreq: cpu:%u, ddi:%u, mem:%u, gpu:%u,\n", curr_clk_tbl->cpu_freq, curr_clk_tbl->ddi_freq, curr_clk_tbl->mem_freq, curr_clk_tbl->gpu_freq);
//	printk("         io :%u, vbu:%u, vco:%u, vol:%d, tbl:%x\n", curr_clk_tbl->io_freq, curr_clk_tbl->vbus_freq, curr_clk_tbl->vcod_freq, new_voltage, tcc_limitclocktbl_flag);
	
	if (new_voltage > curr_voltage) {
		ret = tcc_cpufreq_set_voltage(new_voltage);
		if (ret != 0) {
			pr_err("cpufreq: regulator_set_voltage failed\n");
			return ret;;
		}
	}

	mutex_lock(&mutex_tccnfc);

	if (tcc_freq_old_table.cpu_freq != curr_clk_tbl->cpu_freq) {
		clk_set_rate(pll0_clk, tcc_cpufreq_get_pll_table(curr_clk_tbl->cpu_freq) * 1000);
		clk_set_rate(cpu_clk, curr_clk_tbl->cpu_freq * 1000);
	}
	if (tcc_freq_old_table.mem_freq != curr_clk_tbl->mem_freq)
		clk_set_rate(mem_clk, curr_clk_tbl->mem_freq * 1000);
	if (tcc_freq_old_table.io_freq != curr_clk_tbl->io_freq)
		clk_set_rate(io_clk, curr_clk_tbl->io_freq);
//	if (tcc_freq_old_table.smu_freq != curr_clk_tbl->smu_freq)
//		clk_set_rate(smu_clk, curr_clk_tbl->smu_freq);
	if (tcc_freq_old_table.ddi_freq != curr_clk_tbl->ddi_freq)
		clk_set_rate(ddi_clk, curr_clk_tbl->ddi_freq);
	if (tcc_freq_old_table.gpu_freq != curr_clk_tbl->gpu_freq)
		clk_set_rate(gpu_clk, curr_clk_tbl->gpu_freq);

#if defined(CONFIG_CPU_FREQ_DEFAULT_GOV_PERFORMANCE)
	// Video Thumbnail don't try to create in case both Vcore and Vbus clock is enabled, so Disable VCore clock in case of performance.!! 
	clk_enable(vcod_clk);
#endif
	
	if (tcc_freq_old_table.vcod_freq != curr_clk_tbl->vcod_freq)
		clk_set_rate(vcod_clk, curr_clk_tbl->vcod_freq);
	if (tcc_freq_old_table.vbus_freq != curr_clk_tbl->vbus_freq)
		clk_set_rate(vbus_clk, curr_clk_tbl->vbus_freq);

#if defined(CONFIG_CPU_FREQ_DEFAULT_GOV_PERFORMANCE)
	// Video Thumbnail don't try to create in case both Vcore and Vbus clock is enabled, so Disable VCore clock in case of performance.!! 
	clk_disable(vcod_clk);
#endif

	mutex_unlock(&mutex_tccnfc);

	if (new_voltage < curr_voltage) {
		ret = tcc_cpufreq_set_voltage(new_voltage);
	}

	memcpy(&tcc_freq_old_table, curr_clk_tbl, sizeof(struct tcc_freq_table_t));

	return ret;
}

static int startup_cpufreq = 0;
/* idx: part idx (camera(0), vpu(1), jpge(2), hdmi(3), mali(4), ...)   flag : 1(TRUE) 0(FALSE) */
int tcc_cpufreq_set_limit_table(struct tcc_freq_table_t *limit_tbl, tcc_freq_limit_idx_t idx, int flag)
{
	if (idx >= TCC_FREQ_LIMIT_MAX)
		return -1;

	if (flag) {
		memcpy(&(tcc_freq_limit_table[idx].freq), limit_tbl, sizeof(struct tcc_freq_table_t));
		tcc_freq_limit_table[idx].usecount++;

		tcc_limitclocktbl_flag |= (1<<idx);
		pr_debug("tcc_cpufreq_set_limit_table idx:%d, cnt:%d\n", idx, tcc_freq_limit_table[idx].usecount);
	}
	else {
		#if 1
		memset(&(tcc_freq_limit_table[idx].freq), 0x0, sizeof(struct tcc_freq_table_t));
		tcc_freq_limit_table[idx].usecount = 0;
		#else
		if (tcc_freq_limit_table[idx].usecount == 0)
			return -1;

		if (tcc_freq_limit_table[idx].usecount == 1) {
			memset(&(tcc_freq_limit_table[idx].freq), 0x0, sizeof(struct tcc_freq_table_t));
			tcc_freq_limit_table[idx].usecount--;
		}
		#endif

		tcc_limitclocktbl_flag &= ~(1<<idx);
		pr_debug("tcc_cpufreq_set_limit_table idx:%d, cnt:%d\n", idx, tcc_freq_limit_table[idx].usecount);

		cpufreq_update_policy(0);
	}

	if (flag && startup_cpufreq) {
		if (tcc_freq_limit_table[idx].freq.cpu_freq > tcc_freq_curr_limit_table.cpu_freq)
			tcc_freq_curr_limit_table.cpu_freq = tcc_freq_limit_table[idx].freq.cpu_freq;
		if (tcc_freq_limit_table[idx].freq.ddi_freq > tcc_freq_curr_limit_table.ddi_freq)
			tcc_freq_curr_limit_table.ddi_freq = tcc_freq_limit_table[idx].freq.ddi_freq;
		if (tcc_freq_limit_table[idx].freq.mem_freq > tcc_freq_curr_limit_table.mem_freq)
			tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[idx].freq.mem_freq;
		if (tcc_freq_limit_table[idx].freq.gpu_freq > tcc_freq_curr_limit_table.gpu_freq)
			tcc_freq_curr_limit_table.gpu_freq = tcc_freq_limit_table[idx].freq.gpu_freq;
		if (tcc_freq_limit_table[idx].freq.io_freq > tcc_freq_curr_limit_table.io_freq)
			tcc_freq_curr_limit_table.io_freq = tcc_freq_limit_table[idx].freq.io_freq;
		if (tcc_freq_limit_table[idx].freq.vbus_freq > tcc_freq_curr_limit_table.vbus_freq)
			tcc_freq_curr_limit_table.vbus_freq = tcc_freq_limit_table[idx].freq.vbus_freq;
		if (tcc_freq_limit_table[idx].freq.vcod_freq > tcc_freq_curr_limit_table.vcod_freq)
			tcc_freq_curr_limit_table.vcod_freq = tcc_freq_limit_table[idx].freq.vcod_freq;
		if (tcc_freq_limit_table[idx].freq.smu_freq > tcc_freq_curr_limit_table.smu_freq)
			tcc_freq_curr_limit_table.smu_freq = tcc_freq_limit_table[idx].freq.smu_freq;

		// memory clock fixed rate
		if (tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].usecount && tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].freq.mem_freq)
			tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].freq.mem_freq;
		else if (tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].usecount) {
			if (tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].usecount && tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].freq.mem_freq)
				tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].freq.mem_freq;
			else if (tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].freq.mem_freq)
				tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].freq.mem_freq;
		}

		// cpu clock fixed rate
		if (tcc_freq_limit_table[TCC_FREQ_LIMIT_TV].usecount)
			tcc_freq_curr_limit_table.cpu_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_TV].freq.cpu_freq;

		tcc_cpufreq_set_clock_table(&tcc_freq_curr_limit_table);
	}

	return 0;
}
EXPORT_SYMBOL(tcc_cpufreq_set_limit_table);

static int tcc_cpufreq_target(struct cpufreq_policy *policy,
			      unsigned int target_freq,
			      unsigned int relation)
{
	unsigned int index, i, limit_tbl_flag;
	struct cpufreq_freqs freqs;
	int ret = 0;
	cpufreq_verify_within_limits(policy, policy->cpuinfo.min_freq,
		policy->cpuinfo.max_freq);

	if (target_freq < policy->cpuinfo.min_freq)
		target_freq = policy->cpuinfo.min_freq;
	if (target_freq < policy->min)
		target_freq = policy->min;

	if (tcc_freq_limit_table[TCC_FREQ_LIMIT_FB].usecount == 0) {
		if (policy->cur < 72000 && target_freq >= 364500)
			target_freq = 72000; //144000;
//		else if (policy->cur < 90000 && target_freq >= 364500)
//			target_freq = 90000;
//		else if (policy->cur < 108000 && target_freq >= 364500)
//			target_freq = 108000;
		else if (policy->cur < 126000 && target_freq >= 364500)
			target_freq = 126000;
		else if (policy->cur < 144000 && target_freq >= 364500)
			target_freq = 144000;
		else if (policy->cur < 294000 && target_freq >= 364500)
			target_freq = 294000;
		else if (policy->cur < 364500 && target_freq >= 364500)
			target_freq = 364500;
	}

	if (cpufreq_frequency_table_target(policy, tcc_cpufreq_table,
					   target_freq, relation, &index)) {
		return -EINVAL;
	}

	memcpy(&tcc_freq_curr_limit_table, &(gtClockLimitTable[index]), sizeof(struct tcc_freq_table_t));

	limit_tbl_flag = 0;
	for (i=0 ; i<TCC_FREQ_LIMIT_MAX ; i++) {
		if (tcc_freq_limit_table[i].usecount > 0) {
			if (tcc_freq_limit_table[i].freq.cpu_freq > tcc_freq_curr_limit_table.cpu_freq)
				tcc_freq_curr_limit_table.cpu_freq = tcc_freq_limit_table[i].freq.cpu_freq;
			if (tcc_freq_limit_table[i].freq.ddi_freq > tcc_freq_curr_limit_table.ddi_freq)
				tcc_freq_curr_limit_table.ddi_freq = tcc_freq_limit_table[i].freq.ddi_freq;
			if (tcc_freq_limit_table[i].freq.mem_freq > tcc_freq_curr_limit_table.mem_freq)
				tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[i].freq.mem_freq;
			if (tcc_freq_limit_table[i].freq.gpu_freq > tcc_freq_curr_limit_table.gpu_freq)
				tcc_freq_curr_limit_table.gpu_freq = tcc_freq_limit_table[i].freq.gpu_freq;
			if (tcc_freq_limit_table[i].freq.io_freq > tcc_freq_curr_limit_table.io_freq)
				tcc_freq_curr_limit_table.io_freq = tcc_freq_limit_table[i].freq.io_freq;
			if (tcc_freq_limit_table[i].freq.vbus_freq > tcc_freq_curr_limit_table.vbus_freq)
				tcc_freq_curr_limit_table.vbus_freq = tcc_freq_limit_table[i].freq.vbus_freq;
			if (tcc_freq_limit_table[i].freq.vcod_freq > tcc_freq_curr_limit_table.vcod_freq)
				tcc_freq_curr_limit_table.vcod_freq = tcc_freq_limit_table[i].freq.vcod_freq;
			if (tcc_freq_limit_table[i].freq.smu_freq > tcc_freq_curr_limit_table.smu_freq)
				tcc_freq_curr_limit_table.smu_freq = tcc_freq_limit_table[i].freq.smu_freq;

			limit_tbl_flag |= (1<<i);
		}
	}
	tcc_limitclocktbl_flag = limit_tbl_flag;

	// memory clock fixed rate
	if (tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].usecount && tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].freq.mem_freq)
		tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_HDMI].freq.mem_freq;
	else if (tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].usecount) {
		if (tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].usecount && tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].freq.mem_freq)
			tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_VPU_ENC].freq.mem_freq;
		else if (tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].freq.mem_freq)
			tcc_freq_curr_limit_table.mem_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_CAMERA].freq.mem_freq;
	}

	// cpu clock fixed rate
	if (tcc_freq_limit_table[TCC_FREQ_LIMIT_TV].usecount)
		tcc_freq_curr_limit_table.cpu_freq = tcc_freq_limit_table[TCC_FREQ_LIMIT_TV].freq.cpu_freq;

	freqs.old = policy->cur;
	freqs.new = tcc_freq_curr_limit_table.cpu_freq;
	freqs.cpu = 0;

	cpufreq_notify_transition(&freqs, CPUFREQ_PRECHANGE);
#ifdef CONFIG_REGULATOR
	if (vdd_core == NULL) {
		vdd_core = regulator_get(NULL, "vdd_core");
		if (IS_ERR(vdd_core)) {
			pr_warning("cpufreq: failed to obtain vdd_core\n");
			vdd_core = NULL;
		}
	}
#endif

	pr_debug("cpufreq: changing clk to %u; target = %u, oldfreq = %u, mem = %u, mask = 0x%x\n",
		freqs.new, target_freq, freqs.old, tcc_freq_curr_limit_table.mem_freq, limit_tbl_flag);

	ret = tcc_cpufreq_set_clock_table(&tcc_freq_curr_limit_table);

	startup_cpufreq = 1;

	cpufreq_notify_transition(&freqs, CPUFREQ_POSTCHANGE);

	return ret;
}

static int tcc_cpufreq_verify(struct cpufreq_policy *policy)
{
	if (policy->cpu != 0)
		return -EINVAL;

	return cpufreq_frequency_table_verify(policy, tcc_cpufreq_table);
}

static int tcc_cpufreq_suspend(struct cpufreq_policy *policy, pm_message_t pmsg)
{
	struct tcc_freq_table_t *freqs = &gtClockLimitTable[0];

	clk_set_rate(mem_clk, freqs->mem_freq * 1000);		

	/* Sets minimum CPU frequency and voltage when suspend */
	clk_set_rate(pll0_clk, tcc_cpufreq_get_pll_table(freqs->cpu_freq) * 1000);
	clk_set_rate(cpu_clk, freqs->cpu_freq * 1000);
	clk_set_rate(io_clk, freqs->io_freq);	
//	clk_set_rate(smu_clk, freqs->smu_freq);
	clk_set_rate(ddi_clk, freqs->ddi_freq);
	clk_set_rate(gpu_clk, freqs->gpu_freq);

#if defined(CONFIG_REGULATOR)
	regulator_suspend_prepare(PM_SUSPEND_MEM);
#endif
	return 0;
}

static int tcc_cpufreq_resume(struct cpufreq_policy *policy)
{
	struct tcc_freq_table_t freqs;
	int ret;

	memcpy(&freqs, &gtClockLimitTable[NUM_FREQS - 1], sizeof(struct tcc_freq_table_t));

#if defined(CONFIG_DRAM_MDDR)
	ret = tcc_cpufreq_set_voltage(tcc_cpufreq_get_voltage_by_cpufreq(freqs.cpu_freq));
	if (ret != 0) {
		pr_err("cpufreq: regulator_set_voltage failed\n");
		return -1;
	}

	clk_set_rate(pll0_clk, tcc_cpufreq_get_pll_table(freqs.cpu_freq) * 1000);
	clk_set_rate(cpu_clk, freqs.cpu_freq * 1000);
	clk_set_rate(mem_clk, freqs.mem_freq * 1000);
#elif defined(CONFIG_DRAM_DDR2) || defined(CONFIG_DRAM_DDR3)
	if (freqs.mem_freq >= 250000)
		freqs.cpu_freq = freqs.mem_freq*2;
	
	ret = tcc_cpufreq_set_voltage(tcc_cpufreq_get_voltage_by_cpufreq(freqs.cpu_freq));
	if (ret != 0) {
		pr_err("cpufreq: regulator_set_voltage failed\n");
		return -1;
	}
	
	clk_set_rate(mem_clk, freqs.mem_freq * 1000);
	if (freqs.mem_freq < 250000)
		clk_set_rate(pll0_clk, tcc_cpufreq_get_pll_table(freqs.cpu_freq) * 1000);
	clk_set_rate(cpu_clk, freqs.cpu_freq * 1000);
#else
#error
#endif
	clk_set_rate(io_clk, freqs.io_freq);
//	clk_set_rate(smu_clk, freqs.smu_freq);
	clk_set_rate(ddi_clk, freqs.ddi_freq);
	clk_set_rate(gpu_clk, freqs.gpu_freq);

	memcpy(&tcc_freq_old_table, &freqs, sizeof(struct tcc_freq_table_t));
	return 0;
}

static int __init tcc_cpufreq_init(struct cpufreq_policy *policy)
{
	int i;

	if (policy->cpu != 0)
		return -EINVAL;

	pll0_clk = clk_get(NULL, "pll0");
	if (IS_ERR(pll0_clk))
		return PTR_ERR(pll0_clk);
	cpu_clk = clk_get(NULL, "cpu");
	if (IS_ERR(cpu_clk))
		return PTR_ERR(cpu_clk);
	io_clk = clk_get(NULL, "iobus");
	if (IS_ERR(io_clk))
		return PTR_ERR(io_clk);	
	smu_clk = clk_get(NULL, "smu");
	if (IS_ERR(smu_clk))
		return PTR_ERR(smu_clk);	
	mem_clk = clk_get(NULL, "membus");
	if (IS_ERR(mem_clk))
		return PTR_ERR(mem_clk);
	ddi_clk = clk_get(NULL, "ddi");
	if (IS_ERR(ddi_clk))
		return PTR_ERR(ddi_clk);
	gpu_clk = clk_get(NULL, "gpu");
	if (IS_ERR(gpu_clk))
		return PTR_ERR(gpu_clk);
	vcod_clk = clk_get(NULL, "vcore");
	if (IS_ERR(vcod_clk))
		return PTR_ERR(vcod_clk);
	vbus_clk = clk_get(NULL, "vbus");
	if (IS_ERR(vbus_clk))
		return PTR_ERR(vbus_clk);
	#if defined(CONFIG_USE_ISP)
	isp_bus_clk = clk_get(NULL, "cam");
	if (IS_ERR(isp_bus_clk))
		return PTR_ERR(isp_bus_clk);
	#else
	cam_bus_clk = clk_get(NULL, "cam");
	if (IS_ERR(cam_bus_clk))
		return PTR_ERR(cam_bus_clk);
	#endif

	for (i = 0; i < NUM_FREQS; i++) {
		tcc_cpufreq_table[i].index = i;
		tcc_cpufreq_table[i].frequency = gtClockLimitTable[i].cpu_freq;
	}
	tcc_cpufreq_table[i].frequency = CPUFREQ_TABLE_END;

	policy->cur = policy->min = policy->max = clk_get_rate(cpu_clk);
	policy->cpuinfo.min_freq = tcc_cpufreq_table[0].frequency;
	policy->cpuinfo.max_freq = tcc_cpufreq_table[i - 1].frequency;
	policy->cpuinfo.transition_latency = 
		TCC_TRANSITION_LATENCY * NSEC_PER_USEC;

	cpufreq_frequency_table_cpuinfo(policy, tcc_cpufreq_table);

	printk(KERN_INFO "TCC cpufreq driver initialized\n");
	return 0;
}

static struct cpufreq_driver tcc_cpufreq_driver = {
	.name	= "tcc",
	.owner	= THIS_MODULE,
	.flags	= CPUFREQ_STICKY | CPUFREQ_CONST_LOOPS,
	.init	= tcc_cpufreq_init,
	.verify	= tcc_cpufreq_verify,
	.target	= tcc_cpufreq_target,
	.get	= tcc_cpufreq_get,
	.suspend = tcc_cpufreq_suspend,
	.resume	= tcc_cpufreq_resume,
};

static int __init tcc_cpufreq_register(void)
{
	return cpufreq_register_driver(&tcc_cpufreq_driver);
}

static void __exit tcc_cpufreq_exit(void)
{
	cpufreq_unregister_driver(&tcc_cpufreq_driver);
}

MODULE_AUTHOR("Telechips, Inc.");
MODULE_DESCRIPTION("CPU frequency scaling driver for TCC92xx");

late_initcall(tcc_cpufreq_register);
module_exit(tcc_cpufreq_exit);
