/*
 * linux/drivers/char/tcc_pwrkey.c  
 *
 * Author:  <linux@telechips.com>
 * Created: April 21, 2008
 * Description: Power Key Driver for Telechips Boards.
 *
 * Copyright (C) 2008-2009 Telechips 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * ChangeLog:
 *
 *	Nov, 2008	:	Added TCC79X Suspend and Sleep functions
 *	Oct, 2008	:	Added TCC83X Suspend functions
 *	Apr, 2008	:	Created initialize version
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/delay.h>

#include <linux/interrupt.h>
//#include <asm/arch/irqs.h>
#include <mach/irqs.h>
#include <linux/workqueue.h>

#if defined(CONFIG_ARCH_TCC79X)
//#include <asm/arch/tcc79x.h>
#include <mach/tcc79x.h>
#elif defined(CONFIG_ARCH_TCC83X)
#include <asm/arch/tcc83x_virt.h>
#endif

#define PM_DEV_NAME		"pwr"
#define PM_DEV_MAJOR 	201

#define LONGKEY_TIMEOUT_MSEC	1000

#if defined(CONFIG_ARCH_TCC79X)
#define IRQ_PWRBTN	INT_EI1
#elif defined(CONFIG_ARCH_TCC83X)
#define IRQ_PWRBTN	INT_EI0
#endif

#define LONG_KEY_SUSPEND	0xE1
#define SHORT_KEY_SLEEP		0xE0

extern int plat_tcc_pm(int mode);
extern void plat_tcc_reboot(void);

struct tcc_pm_handle {
	int mode;
	struct work_struct pm_wq;
	int pressed;
	struct timer_list key_timer;
};
static struct tcc_pm_handle pm_h;

static void init_pwrkey(void)
{
#if defined(CONFIG_ARCH_TCC79X)
    BITCLR(HwPORTCFG4, (Hw16-Hw12));			// General Port A3, PowerButton Signal.
    BITCSET(HwCPUD4, (Hw31|Hw30), Hw30);		// Pull Up Enable , 0xF005A110
    BITCSET(HwEINTSEL0, (Hw13-Hw8), (Hw12|Hw8));
    BITCLR(HwGPAEN, Hw3);						// GPA3 input mode

    //Set EINT1 as Level Triggered mode
 	BITSET(HwCLR, Hw1);
	BITCLR(HwMODE, Hw1);						// Edge Triggered mode
	BITSET(HwMODEA, Hw1);						// Both Mode
	BITSET(HwPOL, Hw1);							// Low Active

    while(!(HwGPADAT & Hw3)) {
        printk("HwGPADAT_3 != 1\n");
    }

	BITSET(HwCLR,Hw1);

#elif defined(CONFIG_ARCH_TCC83X)
	BITSET(HwGSEL_B,   HwGSEL_B_N1_EI0 );	// Dedicated function as EINT0
	BITCLR(HwGIOCON_B, Hw4 );				// Port as Input
	BITSET(HwGPIOB_PE, Hw4);				// Pull up|Pull Down Enable
	BITSET(HwGPIOB_PS, Hw4);				// Pull up

	// Interrupt Controller Setting
	BITSET(HwSEL_B,  HwSEL_EI0);			// Normal IRQ
	BITSET(HwCLR_B,  HwCLR_EI0);
	BITSET(HwICFG,   HwICFG_DTYP0_BEdge);	// Both Edge Triggered

	while(!(HwGDATA_B & Hw4)) {
		printk("HwGDATA_B_4 = 1\n");
	}

	BITSET(HwCLR_B,  HwCLR_EI0);
#endif
}

static void enable_pwrkey(void)
{
	// enable pwrkey irq
#if defined(CONFIG_ARCH_TCC79X)
 	BITSET(HwCLR, Hw1);
    BITSET(HwINTMSK, Hw1);
#elif defined(CONFIG_ARCH_TCC83X)
	BITSET(HwCLR_B, HwIEN_EI0); 
	BITSET(HwIEN_B, HwIEN_EI0); 
#endif
}

static void disable_pwrkey(void)
{
	// disable pwrkey irq
#if defined(CONFIG_ARCH_TCC79X)
 	BITSET(HwCLR, Hw1);
    BITCLR(HwINTMSK, Hw1);
#elif defined(CONFIG_ARCH_TCC83X)
	BITSET(HwCLR_B, HwIEN_EI0); 
	BITCLR(HwIEN_B, HwIEN_EI0); 
#endif
}

static void tcc_key_timeout(unsigned long data)
{
	pm_h.mode = LONG_KEY_SUSPEND;
	schedule_work(&(pm_h.pm_wq));
}

static void tcc_wq_func(struct work_struct *wq)
{
	int err;
	
	disable_pwrkey();

	/*
	 * Call Telechips power management function: arch/arm/plat-tcc/tcc_pm.c
	 */
	err = plat_tcc_pm(pm_h.mode);
	if (err) {
		printk("PM %s error.\n", pm_h.mode == SHORT_KEY_SLEEP?"sleep":"suspend");
	} else {
		printk("PM %s success.\n", pm_h.mode == SHORT_KEY_SLEEP?"sleep":"suspend");
	}

	pm_h.pressed = 0;
	enable_pwrkey();
}

static irqreturn_t tcc_pwrkey_handler(int irq, void *dev_id)
{
	if (pm_h.pressed == 0) {
		pm_h.pressed = 1;
		mod_timer(&(pm_h.key_timer), jiffies + msecs_to_jiffies(LONGKEY_TIMEOUT_MSEC));
	} else {
		del_timer(&(pm_h.key_timer));
		pm_h.mode = SHORT_KEY_SLEEP;
		schedule_work(&(pm_h.pm_wq));
	}
	
	return IRQ_HANDLED;
}

static int tcc_pwrkey_ioctl(struct inode *inode, struct file *filp, 
							unsigned int cmd, unsigned long arg)
{
	return 0;
}

static int tcc_pwrkey_release(struct inode *inode, struct file *filp)
{
	return 0;
}

int tcc_pwrkey_open(struct inode *inode, struct file *filp)
{
	return 0;
}

struct file_operations tcc_pwrkey_fops =
{
	.owner		= THIS_MODULE,
	.open		= tcc_pwrkey_open,
	.ioctl		= tcc_pwrkey_ioctl,
	.release	= tcc_pwrkey_release,
};

static int __init pwrkey_init(void)
{
	int res;

	res = register_chrdev(PM_DEV_MAJOR, PM_DEV_NAME, &tcc_pwrkey_fops);
	if (res < 0)
		return res;

	INIT_WORK(&(pm_h.pm_wq), tcc_wq_func);
	pm_h.pressed = 0;

	setup_timer(&(pm_h.key_timer), tcc_key_timeout, (unsigned long)&pm_h);

	init_pwrkey();

	res = request_irq(IRQ_PWRBTN, tcc_pwrkey_handler, IRQF_DISABLED, "pwrkey", NULL);
	if (res)
		return res;

    printk("pwrkey: init\n");
    return 0;
}

static void __exit pwrkey_exit(void)
{
	unregister_chrdev(PM_DEV_MAJOR, PM_DEV_NAME);
	free_irq(IRQ_PWRBTN, NULL);
    printk("pwrkey: exit\n");
}

module_init(pwrkey_init);
module_exit(pwrkey_exit);

MODULE_AUTHOR("Telechips Inc. SYS4-3 linux@telechips.com");
MODULE_DESCRIPTION("TCC Power Button driver");
MODULE_LICENSE("GPL");

