/*
 * linux/drivers/char/tsif/tcc_tsif_parallel.c
 *
 * Author:  <linux@telechips.com>
 * Created: 1st April, 2009
 * Description: Driver for Telechips TS Parallel/Serial Controllers
 *
 * Copyright (c) Telechips, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/dma-mapping.h>
#include <linux/poll.h>
#include <linux/spi/spi.h>

#include <linux/kernel.h>
#include <linux/time.h>
#include <linux/timer.h>
#include <linux/sched.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/atomic.h>

#include <bsp.h>
#include <mach/tcc_pca953x.h>
#include <linux/tcc_tsif_parallel.h>
#include <linux/tcc_tsif_parallel_hwset.h>


struct tcc_tsif_parallel_pri_handle {
    wait_queue_head_t wait_q;
    struct mutex mutex;
    int open_cnt;
};


static struct tcc_tsif_parallel_handle tsif_handle;
static struct tcc_tsif_parallel_pri_handle tsif_pri;

#if 0
static int __init tsif_drv_probe(struct platform_device *dev)
{
    platform_set_drvdata(dev, &tsif_handle);
    return 0;
}
#endif

#ifdef CONFIG_PM
/*
 * NOTE: before suspend, you must close spi.
 */
#if 0
static int tsif_drv_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;
}
static int tsif_drv_resume(struct platform_device *dev)
{
	return 0;
}
#endif
#else
#define tsif_drv_suspend NULL
#define tsif_drv_resume  NULL
#endif

#if 0
static struct platform_driver tsif_platform_driver = {
	.driver = {
		.name	= "tcc-tsif",
		.owner	= THIS_MODULE,
	},
	.suspend = tsif_drv_suspend,
	.resume  = tsif_drv_resume,
};
#endif

static int tsif_parallel_dxb_power(unsigned long flag)
{
	if(flag & 0x10)
	{
		if (flag == 1) {
			/* GPIO_EXPAND CAM_ON Power-on */
			tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, CAM, OUTPUT, HIGH, SET_DIRECTION | SET_VALUE);
		} else if (flag == 0) {
			/* GPIO_EXPAND DXB_ON Power-off */
			tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, CAM, OUTPUT,  LOW, SET_DIRECTION | SET_VALUE);
		} else {
			return -EINVAL;
		}
		printk("[%s %d]: CAM_ON port %s\n", __func__, __LINE__, (tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, 0, OUTPUT, 0, GET_VALUE)&CAM)?"ON":"OFF");
		return 0;
	}
	else
	{
		if (flag == 1) {
			/* GPIO_EXPAND DXB_ON Power-on */
			tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, DXB, OUTPUT, HIGH, SET_DIRECTION|SET_VALUE);
		} else if (flag == 0) {
			/* GPIO_EXPAND DXB_ON Power-off */
			tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, DXB, OUTPUT, LOW, SET_DIRECTION|SET_VALUE);
		} else {
			return -EINVAL;
		}
		printk("tsif_parallel: DXB_ON port %s\n", (tcc_pca953x_setup(PCA9539_U3_SLAVE_ADDR, 0, OUTPUT, 0, GET_VALUE)&DXB)?"ON":"OFF");
		return 0;
	}
}


static void tea_free_dma_linux(struct tea_dma_buf *tdma)
{
    if(tdma) {
        if(tdma->v_addr != 0) {
            dma_free_writecombine(0, tdma->buf_size, tdma->v_addr, tdma->dma_addr);

			printk(KERN_INFO "tcc-tsif-parallel : dma buffer free @0x%X(Phy=0x%X), size:%d\n", (unsigned int)tdma->v_addr, (unsigned int)tdma->dma_addr, tdma->buf_size);
        }
        memset(tdma, 0, sizeof(struct tea_dma_buf));
    }
}

static int tea_alloc_dma_linux(struct tea_dma_buf *tdma, unsigned int size)
{
    int ret = -1;

    if(tdma) {
        tea_free_dma_linux(tdma);

        tdma->buf_size = size;
        tdma->v_addr = dma_alloc_writecombine(0, tdma->buf_size, &tdma->dma_addr, GFP_KERNEL);

        ret = tdma->v_addr ? 0 : 1;

		printk(KERN_INFO "tcc-tsif-parallel : dma buffer alloc @0x%X(Phy=0x%X), size:%d\n", (unsigned int)tdma->v_addr, (unsigned int)tdma->dma_addr, tdma->buf_size);
    }

    return ret;
} 

static int tsif_get_readable_cnt(struct tcc_tsif_parallel_handle *h)
{
	int ret = 0;
	
	if(h){
		if(h->dma_recv_size > 0)	ret = h->dma_recv_size / MPEG_PACKET_SIZE;
			
		return ret;
	}

    return ret;
}

static irqreturn_t tsif_parallel_dma_handler(int irq, void *dev_id)
{
	struct tcc_tsif_parallel_handle *tspi 	= (struct tcc_tsif_parallel_handle *)dev_id;
    struct tcc_tsif_parallel_pri_handle *tpri = (struct tcc_tsif_parallel_pri_handle *)tspi->private_data;
    
	tsif_handle.tsif_isr(tspi); 
	
	if(tsif_get_readable_cnt(tspi) > 0) {
		wake_up(&(tpri->wait_q));
	}

    return IRQ_HANDLED;
}

static ssize_t tsif_parallel_read(struct file *filp, char *buf, size_t len, loff_t *ppos)
{
	int cnt;
	volatile PGDMACTRL dma_regs;
	
	dma_regs = (volatile PGDMACTRL)tcc_p2v(HwGDMA0_BASE); 

	tsif_handle.dma_phy_curpos = dma_regs->C_DADR0;
	
	if(tsif_handle.dma_phy_curpos > tsif_handle.dma_phy_rdpos) {
		tsif_handle.dma_recv_size = tsif_handle.dma_phy_curpos - tsif_handle.dma_phy_rdpos;
	} else { 
		tsif_handle.dma_recv_size = (tsif_handle.dma_phy_maxaddr - tsif_handle.dma_phy_rdpos) + (tsif_handle.dma_phy_curpos - tsif_handle.rx_dma.dma_addr);
	}
	
	cnt = tsif_get_readable_cnt(&tsif_handle); 
	
	if(cnt) {
		int offset;
		int copy_size;
		
		if((cnt * MPEG_PACKET_SIZE) < len) 		copy_size = cnt * MPEG_PACKET_SIZE;
		else									copy_size = len;
		
		if((tsif_handle.dma_phy_rdpos + copy_size) < tsif_handle.dma_phy_maxaddr) {
			offset = tsif_handle.dma_phy_rdpos - tsif_handle.rx_dma.dma_addr;
			
			if(copy_to_user(buf, (tsif_handle.rx_dma.v_addr + offset), copy_size)) {
				return -EFAULT;
			}
		} else {
			int first_size;
			int second_size;
			
			offset = tsif_handle.dma_phy_rdpos - tsif_handle.rx_dma.dma_addr;
			
			first_size 	= tsif_handle.dma_phy_maxaddr - tsif_handle.dma_phy_rdpos;
			if(copy_to_user(buf, (tsif_handle.rx_dma.v_addr + offset), first_size)) {
				return -EFAULT;
			}
			
			second_size	= copy_size - first_size;
			if(copy_to_user((buf + first_size), tsif_handle.rx_dma.v_addr, second_size)) {
				return -EFAULT;
			}
		}
		
		if(tsif_handle.dma_phy_curpos > tsif_handle.dma_phy_rdpos) {
			tsif_handle.dma_phy_rdpos = tsif_handle.dma_phy_curpos - ((tsif_handle.dma_phy_curpos - tsif_handle.dma_phy_rdpos) % MPEG_PACKET_SIZE);
		} else {
			tsif_handle.dma_phy_rdpos = (tsif_handle.dma_phy_curpos - ((tsif_handle.dma_phy_maxaddr - tsif_handle.dma_phy_rdpos + (tsif_handle.dma_phy_curpos - tsif_handle.rx_dma.dma_addr)) % MPEG_PACKET_SIZE));
		}
		
		if(tsif_handle.dma_intr_cnt > 1)	printk("%d\n", tsif_handle.dma_intr_cnt);
		
		tsif_handle.dma_recv_size 	= 0;
		tsif_handle.dma_intr_cnt	= 0;
		
		return copy_size;
	}

	return 0;
}

static ssize_t tsif_parallel_write(struct file *filp, const char *buf, size_t len, loff_t *ppos)
{
    return 0;
}

static unsigned int tsif_parallel_poll(struct file *filp, struct poll_table_struct *wait)
{
    if(tsif_get_readable_cnt(&tsif_handle) > 0) {
		return  (POLLIN | POLLRDNORM);
    } 
	
	poll_wait(filp, &(tsif_pri.wait_q), wait);

    if(tsif_get_readable_cnt(&tsif_handle) > 0) {
		return  (POLLIN | POLLRDNORM);
    }

    return 0;
}

static int tsif_parallel_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	switch (cmd) {
		case IOCTL_TSIF_DMA_START:
			{
				if(copy_from_user(&tsif_handle.msb_first, (void *)arg, 6)) {
					printk(KERN_ERR "cannot copy from user tsif_parallel_param in IOCTL_TSIF_DMA_START !!! \n");
					return -EFAULT;
				}

				tsif_handle.dma_stop(&tsif_handle);
				tsif_handle.tsif_set(&tsif_handle);
				tsif_handle.clear_fifo_packet(&tsif_handle);
				tsif_handle.dma_start(&tsif_handle);
			}
			break;
			
		case IOCTL_TSIF_DMA_STOP:
	            tsif_handle.dma_stop(&tsif_handle);
    	    break;
			
		case IOCTL_TSIF_GET_MAX_DMA_SIZE:
			{
	            struct tcc_tsif_parallel_param param;

	            param.ts_total_packet_cnt = tsif_handle.dma_total_size / TSIF_PACKET_SIZE;
	            param.ts_intr_packet_cnt = 1;

	            if (copy_to_user((void *)arg, (void *)&param, sizeof(struct tcc_tsif_parallel_param))) {
	                printk("cannot copy to user tsif_parallel_param in IOCTL_TSIF_GET_MAX_DMA_SIZE !!! \n");
	                return -EFAULT;
	            }
	        }
	        break;
			
		case IOCTL_TSIF_DXB_POWER:
			{
				unsigned long flag;
				if (copy_from_user((void *)&flag, (const void *)arg, sizeof(flag))) {
					return -EFAULT;
				}
				if (tsif_parallel_dxb_power(flag)) {
					return -EINVAL;
				}
			}
			break;

		case IOCTL_TSIF_SET_PID:
	    default:
	        printk(KERN_INFO "tcc-tsif-parallel : unrecognized ioctl (0x%X)\n", cmd);
	        ret = -EINVAL;
	        break;
    }
    return ret;
}

static int tsif_parallel_init(int tsif_num)
{
	int ret = 0;
    volatile struct tcc_tsif_regs *reg_addr = NULL;

    reg_addr = (volatile struct tcc_tsif_regs *)((TSIF_CH == 0) ? &(HwTSIF0_BASE) : &(HwTSIF1_BASE));

    memset(&tsif_handle, 0, sizeof(struct tcc_tsif_parallel_handle));

	tca_ckc_set_iobus_swreset(RB_TSIFCONTROLLER, OFF);	//ahb tsif sw reset
	tca_ckc_set_iobus_swreset(RB_TSIFCONTROLLER, ON);

	tca_ckc_setiobus(RB_TSIFCONTROLLER, ENABLE);		//ahb tsif clock enable
	
	tsif_handle.dma_intr_packet_cnt = 512;				//TSIF interrput is occur when device is gather the amount of data until dma_intr_packet_cnt * 188

	if(tcc_tsif_parallel_init(&tsif_handle,
					reg_addr,
					tea_alloc_dma_linux,
					tea_free_dma_linux,
					TSIF_DMA_SIZE,
					TSIF_CH,
					TSIF_DMA_CONTROLLER,
					TSIF_DMA_CH))
	{

        printk(KERN_ERR "tcc-tsif-parallel : %s : tcc_tsif_parallel_init error !!!!!\n", __func__);

		ret = -EBUSY;
		goto err_tsif;
    }

	tsif_handle.private_data = (void *)&tsif_pri;
	
	ret = request_irq(tsif_handle.irq, tsif_parallel_dma_handler, IRQF_SHARED, TCC_TSIF_PARALLEL_DEV_NAME, &tsif_handle);
	if(ret) { 
		goto err_irq; 
	}

	return 0;

err_irq:
	free_irq(tsif_handle.irq, &tsif_handle);

err_tsif:
	tcc_tsif_parallel_clean(&tsif_handle);
	tca_ckc_setiobus(RB_TSIFCONTROLLER, DISABLE);		//ahb tsif clock disable

	return ret;
}

static void tsif_parallel_deinit(int tsif_num)
{
	free_irq(tsif_handle.irq, &tsif_handle);

	tsif_handle.dma_stop(&tsif_handle);
	
    tcc_tsif_parallel_clean(&tsif_handle);
	tca_ckc_setiobus(RB_TSIFCONTROLLER, DISABLE);		//ahb tsif clock disable
}


static int tsif_parallel_open(struct inode *inode, struct file *filp)
{
    int ret = 0;
	
    mutex_lock(&(tsif_pri.mutex));
	
    if(tsif_pri.open_cnt == 0) {
        tsif_pri.open_cnt++;
		
		ret = tsif_parallel_init(TSIF_CH);
    } else {
        ret = -EBUSY;
    }

    mutex_unlock(&(tsif_pri.mutex));
    return ret;
}


static int tsif_parallel_release(struct inode *inode, struct file *filp)
{
    mutex_lock(&(tsif_pri.mutex));
	
    if(tsif_pri.open_cnt > 0) {
        tsif_pri.open_cnt--;
		tsif_parallel_deinit(TSIF_CH);
    }

    mutex_unlock(&(tsif_pri.mutex));
    return 0;
}


struct file_operations tcc_tsif_parallel_fops = {
    .owner      = THIS_MODULE,
    .read       = tsif_parallel_read,
    .write      = tsif_parallel_write,
    .ioctl      = tsif_parallel_ioctl,
    .open       = tsif_parallel_open,
    .release    = tsif_parallel_release,
    .poll       = tsif_parallel_poll,
};


static int __init init_tsif_parallel_dev(void)
{
    int ret;

	printk(KERN_INFO "tcc_tsif_parallel_init : dev-name : %s, tsif%d & dma0-%d use\n", TCC_TSIF_PARALLEL_DEV_NAME, TSIF_CH, TSIF_DMA_CH);

	memset(&tsif_pri, 0, sizeof(struct tcc_tsif_parallel_pri_handle));
	ret = register_chrdev(TCC_TSIF_PARALLEL_DEV_MAJOR, TCC_TSIF_PARALLEL_DEV_NAME, &tcc_tsif_parallel_fops);
	if(ret < 0) {
		printk(KERN_ERR "tcc_tsif_parallel_init [%s:%d] register_chrdev error !!!!!\n", __func__, __LINE__); 
		return ret;
	} 

    init_waitqueue_head(&(tsif_pri.wait_q));
    mutex_init(&(tsif_pri.mutex));

	//ret = platform_driver_probe(&tsif_platform_driver, tsif_drv_probe);
	
    return ret;
}


static void __exit exit_tsif_parallel_dev(void)
{
    unregister_chrdev(TCC_TSIF_PARALLEL_DEV_MAJOR, TCC_TSIF_PARALLEL_DEV_NAME);

	//platform_driver_unregister(&tsif_platform_driver);
}



module_init(init_tsif_parallel_dev);
module_exit(exit_tsif_parallel_dev);

MODULE_AUTHOR("Telechips Inc. SYS4-5 linux@telechips.com");
MODULE_DESCRIPTION("Telechips TSIF driver");
MODULE_LICENSE("GPL");

