/*
 * linux/drivers/i2c/chips/tcc_pca953x.c
 *
 * Author:  <linux@telechips.com>
 * Created: 31th March, 2009 
 * Description: Tcc250 Driver 
 *
 * Copyright (c) Telechips, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ====================================
 *  Description
 *  -----------
 * Slave chip: two PCA9539, one PCA9538
 * Slave address: PCA9539_U2: 0x77	// 1110 1HH[R/W]
 *                PCA9539_U3: 0x74	// 1110 1LL[R/W]
 *                PCA9538_U4: 0x70	// 1110 0LL[R/W]
 * EXPORT_SYMBOL(tcc_pca953x_read);
 * EXPORT_SYMBOL(tcc_pca953x_write);
 * EXPORT_SYMBOL(tcc_pca953x_setup);
 * ====================================
 *
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <mach/tcc_pca953x.h>


#define DEVICE_NAME		"pca953x"
#define MAJOR_ID		210
#define MINOR_ID		1

static unsigned short probe_i2c[] = {0, PCA9539_U2_SLAVE_ADDR, I2C_CLIENT_END};	/* { i2c_num, i2c_addr } */
static unsigned short dummy[] = {I2C_CLIENT_END};

static struct i2c_client_address_data addr_data = {
	.normal_i2c = dummy,
	.probe = probe_i2c,
	.ignore = dummy,
};

static struct i2c_driver pca953x_i2c_driver;
struct i2c_client *pca953x_i2c_client;
//EXPORT_SYMBOL(pca953x_i2c_client);

static int pca953x_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *i2c;

	i2c = kzalloc(sizeof(*i2c), GFP_KERNEL);
	if (!i2c)
		return -ENOMEM;

	strcpy(i2c->name, "PCA953X");
	i2c->flags = 0;
	i2c->addr = addr;
	i2c->adapter = adap;
	i2c->driver = &pca953x_i2c_driver;

	pca953x_i2c_client = i2c;

	return i2c_attach_client(i2c);
}

static int pca953x_i2c_detach(struct i2c_client *client)
{
	i2c_detach_client(client);
	kfree(pca953x_i2c_client);
	return 0;
}

static int pca953x_i2c_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &addr_data, pca953x_i2c_probe);
}

static struct i2c_driver pca953x_i2c_driver = {
	.driver = {
		.name = "PCA953X",
		.owner = THIS_MODULE,
	},
	.id = 1,
	.attach_adapter = pca953x_i2c_attach,
	.detach_client = pca953x_i2c_detach,
};

/*------------------------------------------------------------------------
 * PCA953X Control APIs
 *
 *------------------------------------------------------------------------
 */
int tcc_pca953x_read(int slave, unsigned char cmd, unsigned char *rd_buf)
{
	struct i2c_adapter *adap = pca953x_i2c_client->adapter;
	struct i2c_msg msgs[2];
	int ret;
		
	/* write port_nr */
	msgs[0].addr = slave;
	msgs[0].flags = pca953x_i2c_client->flags & I2C_M_TEN;
	msgs[0].len = 1;
	msgs[0].buf = &cmd;

	/* read port */
	msgs[1].addr = slave;
	msgs[1].flags = pca953x_i2c_client->flags & I2C_M_TEN;
	msgs[1].flags |= I2C_M_RD;
	msgs[1].len = 1;
	msgs[1].buf = rd_buf;

	ret = i2c_transfer(adap, msgs, 2);
	if (ret != 1) {
		return -1;
	}

	return 0;
}
EXPORT_SYMBOL(tcc_pca953x_read);

int tcc_pca953x_write(int slave, const unsigned char *wr_buf, int count)
{
	struct i2c_adapter *adap = pca953x_i2c_client->adapter;
	struct i2c_msg msgs;
	int ret;

	msgs.addr = slave;
	msgs.flags = pca953x_i2c_client->flags & I2C_M_TEN;
	msgs.len = count;
	msgs.buf = (__u8 *)wr_buf;
	ret = i2c_transfer(adap, &msgs, 1);
	if (ret != 1) {
		return -1;
	}

	return 0;
}
EXPORT_SYMBOL(tcc_pca953x_write);

int tcc_pca953x_setup(int slave, int name, int direction, int value, int mode)
{
	int ret = 0, port_nr = 0, port[2], conf[2];
	unsigned char buf[2];

	switch (slave) {
		case PCA9539_U2_SLAVE_ADDR:
			port_nr = 2;
			port[0] = direction ? PCA9539_INPUT_0 : PCA9539_OUTPUT_0;
			port[1] = direction ? PCA9539_INPUT_1 : PCA9539_OUTPUT_1;
			conf[0] = PCA9539_DIRECTION_0;
			conf[1] = PCA9539_DIRECTION_1;
			break;
			
		case PCA9539_U3_SLAVE_ADDR:
			port_nr = 2;
			port[0] = direction ? PCA9539_INPUT_0 : PCA9539_OUTPUT_0;
			port[1] = direction ? PCA9539_INPUT_1 : PCA9539_OUTPUT_1;
			conf[0] = PCA9539_DIRECTION_0;
			conf[1] = PCA9539_DIRECTION_1;
			break;
			
		case PCA9538_U4_SLAVE_ADDR:
			port_nr = 1;
			port[0] = direction ? PCA9538_INPUT_0 : PCA9538_OUTPUT_0;
			conf[0] = PCA9538_DIRECTION_0;
			break;
			
		default:
			break;
	}

	/*
	 * GET_VALUE is read only
	 */
	if (GET_VALUE == mode) {
		ret |= tcc_pca953x_read(slave, port[0], &buf[0]);
		if (port_nr == 2) {
			ret |= tcc_pca953x_read(slave, port[1], &buf[1]);
		} else {
			buf[1] = 0;
		}
		
		if (ret) ret = -1;
		else ret = ((buf[1] << 8) | buf[0]) & 0x0000ffff;
		return ret;
	}

	/*
	 * What port number?
	 */
	if (name >= Hw0 && name <= Hw7) {
		port_nr = 0;
	} else if (name >= Hw8 && name <= Hw15) {
		port_nr = 1;
		name >>= 8;
	} else {
		return -1;
	}

	/*
	 * Set direction
	 */
	if (SET_DIRECTION & mode) {
		tcc_pca953x_read(slave, conf[port_nr], &buf[0]);
		if (INPUT == direction) {
			buf[1] = buf[0] | name;
		} else if (OUTPUT == direction) {
			buf[1] = buf[0] & ~name;
		} else {
			return -1;
		}
		
		buf[0] = conf[port_nr];
		ret = tcc_pca953x_write(slave, &buf[0], 2);
		if (ret == -1)
			return -1;
	}

	/*
	 * Set value
	 */
	if (SET_VALUE & mode) {
		tcc_pca953x_read(slave, port[port_nr], &buf[0]);
		if (HIGH == value) {
			buf[1] = buf[0] | name;
		} else if (LOW == value) {
			buf[1] = buf[0] & ~name;
		} else {
			return -1;
		}
		
		buf[0] = port[port_nr];
		tcc_pca953x_write(slave, &buf[0], 2);
		if (ret == -1)
			return -1;
	}

	return 0;
}
EXPORT_SYMBOL(tcc_pca953x_setup);

static int tcc_pca953x_ioctl(struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
	return 0;
}

static int tcc_pca953x_open(struct inode *inode, struct file *filp)
{	
	return 0;
}

static int tcc_pca953x_release(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations tcc_pca953x_fops = 
{
	.owner			= THIS_MODULE,
	.ioctl			= tcc_pca953x_ioctl,
	.open			= tcc_pca953x_open,
	.release		= tcc_pca953x_release,	
};

static struct class *tcc_pca953x_class;

static int __init tcc_pca953x_i2c_init(void)
{
	register_chrdev(MAJOR_ID, DEVICE_NAME, &tcc_pca953x_fops);
	tcc_pca953x_class = class_create(THIS_MODULE, DEVICE_NAME);
	device_create(tcc_pca953x_class, NULL, MKDEV(MAJOR_ID, MINOR_ID), NULL, DEVICE_NAME);

	return i2c_add_driver(&pca953x_i2c_driver);
}

static void __exit tcc_pca953x_i2c_exit(void)
{
	unregister_chrdev(MAJOR_ID, DEVICE_NAME);

	i2c_del_driver(&pca953x_i2c_driver);
}

module_init(tcc_pca953x_i2c_init);
module_exit(tcc_pca953x_i2c_exit);

MODULE_AUTHOR("Telechips Inc. SYS4-3 linux@telechips.com");
MODULE_DESCRIPTION("PCA953x I2C driver");
MODULE_LICENSE("GPL");

