/* linux/drivers/mtd/nand/tcc79xx_mtd_nand.c
 *
 * Copyright (c) 2004,2005 Simtec Electronics
 *	http://www.simtec.co.uk/products/SWLINUX/
 *	Ben Dooks <ben@simtec.co.uk>
 *
 * Telechips TCC7901 NAND driver
 *
 * Changelog:
 *	10-Jan-2009  JWK  Initial version
 *
 * $Id: TCC7901.c,v 1.23 2006/04/01 18:06:29 bjd Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/slab.h>
#include <linux/clk.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>

#include <mach/tcc79x.h>


/**********************************************************
*	void IO_CKC_EnableBUS(unsigned X);
*	Input		: X = Bitmap dedicated to each peripheral
*	Return		:
*	Description	: Enable(1) the bus clock fed to each peripheral.
*				  The other bus clocks are not influenced.
**********************************************************/
#define	IO_CKC_EnableBUS(X)			(BITSET(HwBCLKCTR, (X)))
/**********************************************************
*	void IO_CKC_DisableBUS(unsigned X);
*	Input		: X = Bitmap dedicated to each peripheral
*	Return		:
*	Description	: Disable(1) the bus clock fed to each peripheral.
*				  The other bus clocks are not influenced.
**********************************************************/
#define	IO_CKC_DisableBUS(X)			(BITCLR(HwBCLKCTR, X))


void static inline tcc_writeb(u8 val, unsigned long reg)
{
	*(volatile u8 *)(reg) = val;
}

void static inline tcc_writew(u16 val, unsigned long reg)
{
	*(volatile u16 *)(reg) = val;
}

void static inline tcc_writel(u32 val, unsigned long reg)
{
	*(volatile u32 *)(reg) = val;
}

static inline u8 tcc_readb(unsigned long reg)
{
	return (*(volatile u8 *)reg);
}

static inline u16 tcc_readw(unsigned long reg)
{
	return (*(volatile u16 *)reg);
}

static inline u32 tcc_readl(unsigned long reg)
{
	return (*(volatile u32 *)reg);
}

/*
 * MTD structure for NAND controller
 */
static struct mtd_info *tcc7901_mtd = NULL;

/*
 * Define partitions for flash device
 */
static const struct mtd_partition partition_info[] = {
/*  
	{
	 .name = "Bootloader",
	 .offset = 0,
	 .size = 21 * 1024 * 1024},
*/	
/* 
	{
	 .name = "bad",
	 .offset = 30 * 1024 * 1024,
	 .size = 120 * 1024 * 1024},	
*/	 
	{
	 .name = "system",
	 .offset = 160 * 1024 * 1024,
	 .size = 120 * 1024 * 1024},		 
	{
	 .name = "cache",
	 .offset = MTDPART_OFS_APPEND,
	 .size = 60 * 1024 * 1024},			 
	{
	 .name = "userdata",
	 .offset = MTDPART_OFS_APPEND,
	 .size = 200 * 1024 * 1024}	 
/*	 
	{
	 .name = "userdata",
	 .offset = MTDPART_OFS_APPEND,
	 .size = MTDPART_SIZ_FULL}
*/	 
};

/**
 * tcc_read_byte -  read one byte from the chip
 * @mtd:	MTD device structure
 *
 *  read function for 8bit buswith
 */
static u_char tcc_read_byte(struct mtd_info *mtd)
{
	u_char ret = HwNFC_SDATA & 0x000000FF;
	return ret;
}

/**
 * tcc_write_buf -  write buffer to chip
 * @mtd:	MTD device structure
 * @buf:	data buffer
 * @len:	number of bytes to write
 *
 *  write function for 8bit buswith
 */
static void tcc_write_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	int i;
	for (i = 0; i < len; i++) {
		HwNFC_SDATA = 0x000000FF & buf[i];
	}
}

/**
 * tcc_read_buf -  read chip data into buffer
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 *  read function for 8bit buswith
 */
static void tcc_read_buf(struct mtd_info *mtd, u_char *buf, int len)
{
	int i;
	for (i = 0; i < len; i++) {
		buf[i] = HwNFC_SDATA & 0x000000FF;
	}
}

/**
 * tcc_verify_buf -  Verify chip data against buffer
 * @mtd:	MTD device structure
 * @buf:	buffer containing the data to compare
 * @len:	number of bytes to compare
 *
 *  verify function for 8bit buswith
 */
static int tcc_verify_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	int i;
	for (i = 0; i < len; i++) {
		if (buf[i] != (HwNFC_SDATA & 0x000000FF))
			return -EFAULT;
	}

	return 0;
}



int tcc7901_device_ready(struct mtd_info *mtd)
{
	//wkjung
	int ret = ISZERO( HwNFC_CTRL, HwNFC_CTRL_RDY_RDY );
	
	if (ret == 1) ret = 0;
	else if (ret == 0) ret = 1;
	
	return ret;
}

/**
 * tcc7901_select_chip - control -CE line
 *	Forbid driving -CE manually permitting the NAND controller to do this.
 *	Keeping -CE asserted during the whole sector reads interferes with the
 *	NOR flash and PCMCIA drivers as it causes contention on the static bus.
 *	We only have to hold -CE low for the NAND read commands since the flash
 *	chip needs it to be asserted during chip not ready time but the NAND
 *	controller keeps it released.
 *
 * @mtd:	MTD device structure
 * @chip:	chipnumber to select, -1 for deselect
 */
#define ASM_NOP {					\
	__asm__ __volatile__ ("nop");	\
	__asm__ __volatile__ ("nop");	\
	__asm__ __volatile__ ("nop");	\
	__asm__ __volatile__ ("nop");	\
	__asm__ __volatile__ ("nop");	\
	__asm__ __volatile__ ("nop");	\
} 

static void tcc7901_select_chip(struct mtd_info *mtd, int chipnr)
{
	struct nand_chip *chip = mtd->priv;

	switch (chipnr) {
	case -1:
		HwGPFDAT	&= ~Hw6;		/* Set Low */
		chip->cmd_ctrl(mtd, NAND_CMD_NONE, 0 | NAND_CTRL_CHANGE);		
		break;
	case 0:
		HwGPFDAT	|= Hw6;		/* Set High */
		break;

	default:
		BUG();
	}
	
return;
	
}

/* tcc7901_nand_hwcontrol
 *
 * Issue command and address cycles to the chip
*/
static void tcc7901_nand_hwcontrol(struct mtd_info *mtd, int cmd,
				   unsigned int ctrl)
{
	if (cmd == NAND_CMD_NONE) {
		return;
	}

	if (ctrl & NAND_CLE)
		HwNFC_CMD = 0x000000FF & cmd;		
	else
		HwNFC_SADDR = 0x000000FF & cmd;
}

/*
 * Main initialization routine
 */
static int __init tcc7901_nand_init(void)
{
	struct nand_chip *this;
	int retval;

	HwGPFDAT	&= ~Hw6;		/* Set Low */
	HwGPFEN		|= Hw6;		/* OUTPUT MODE */
	HwPORTCFG3	|= (1)*Hw28;		/* Normal IO Port */

	IO_CKC_EnableBUS( HwBCLKCTR_ND_ON );

	/* Make Reset */
	HwNFC_RST	= 0;

	/* Set Default NFC Configuration */
	HwNFC_CTRL	= HwNFC_CTRL_DEN_EN		|
			      HwNFC_CTRL_CFG_NOACT	|
			      HwNFC_CTRL_BSIZE_1	|
				      (4 << 4)				|		// pw = 5
			      (1 << 0);						// hold = 1

	/* Set Setup Time and Hold Time */
	BITCSET( HwNFC_CTRL, 0xFFF, 0x466 );

	/* Enable Chip Select */
	BITSCLR( HwNFC_CTRL, HwNFC_CTRL_CFG_NOACT, HwNFC_CTRL_CFG_nCS0 );
 	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
 	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP;
	ASM_NOP;ASM_NOP;ASM_NOP;ASM_NOP; 

	/* Set Data Bus as 16Bit */
	BITSET( HwNFC_CTRL, HwNFC_CTRL_BW_16 );
	

	/* Allocate memory for MTD device structure and private data */
	tcc7901_mtd = kmalloc(sizeof(struct mtd_info) + sizeof(struct nand_chip), GFP_KERNEL);
	if (!tcc7901_mtd) {
		printk("Unable to allocate NAND MTD dev structure.\n");
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *)(&tcc7901_mtd[1]);

	/* Initialize structures */
	memset(tcc7901_mtd, 0, sizeof(struct mtd_info));
	memset(this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	tcc7901_mtd->priv = this;
	tcc7901_mtd->owner = THIS_MODULE;


	/* Set address of hardware control function */
	this->dev_ready = tcc7901_device_ready;
	this->select_chip = tcc7901_select_chip;
	this->cmd_ctrl  = tcc7901_nand_hwcontrol;

	/* 30 us command delay time */
	this->chip_delay = 30;
	this->ecc.mode = NAND_ECC_NONE;

	this->read_byte = tcc_read_byte;
	this->write_buf = tcc_write_buf;
	this->read_buf = tcc_read_buf;
	this->verify_buf = tcc_verify_buf;
	this->IO_ADDR_R = this->IO_ADDR_W = (void *) HwNFC_SDATA;
	/* Scan to find existence of the device */
	if (nand_scan(tcc7901_mtd, 1)) {
		retval = -ENXIO;
		goto outmem;
	}

	/* Register the partitions */
	add_mtd_partitions(tcc7901_mtd, partition_info, ARRAY_SIZE(partition_info));

	return 0;

 outmem:
	kfree(tcc7901_mtd);
	return retval;
}

module_init(tcc7901_nand_init);

/*
 * Clean up routine
 */
static void __exit tcc7901_cleanup(void)
{
	/* Release resources, unregister device */
	nand_release(tcc7901_mtd);

	/* Free the MTD device structure */
	kfree(tcc7901_mtd);

	IO_CKC_DisableBUS( HwBCLKCTR_ND_ON );
}

module_exit(tcc7901_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Telechips, JWK");
MODULE_DESCRIPTION("Board-specific glue layer for NAND flash on Tcc79xx board");

