/*
 * linux/drivers/serial/rtc-tcc.c
 *
 * Author: <linux@telechips.com>
 * Created: June 10, 2008
 * Description: RTC driver for Telechips TCC Series
 *
 * Copyright (C) 2008-2009 Telechips 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <linux/clk.h>

#include <mach/hardware.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/rtc.h>
#include <asm/delay.h>
#include <asm/mach/time.h>

#ifdef CONFIG_ARCH_TCC79X
#include <mach/tcc79x.h>
#elif CONFIG_ARCH_TCC83X
#include <mach/tcc83x_virt.h>
#endif

#define DRV_NAME "tcc-rtc"

#include "tca_alarm.h"

static int tcc_rtc_alarmno = NO_IRQ;

//#define pr_debug printk

/* IRQ Handlers */
static irqreturn_t tcc_rtc_alarmirq(int irq, void *id)
{
    //printk("%s: IRQ OK\n", __func__);

    tca_alarm_setint();

	rtc_update_irq(id, 1, RTC_AF | RTC_IRQF);

	return IRQ_HANDLED;
}


/* Update control registers */
static void tcc_rtc_setaie(int to)
{
	unsigned int tmp;

	pr_debug("%s: aie=%d\n", __func__, to);

	tmp = HwRTCALM & ~HwRTCALM_ALMEN_EN;

	if (to)
		tmp |= HwRTCALM_ALMEN_EN;

	HwRTCALM = tmp;	
}


static int tcc_rtc_gettime(struct device *dev, struct rtc_time *rtc_tm)
{
    rtctime pTime;

    local_irq_disable();
    tca_rtc_gettime(&pTime);

	rtc_tm->tm_sec	= pTime.wSecond;
	rtc_tm->tm_min	= pTime.wMinute;
	rtc_tm->tm_hour	= pTime.wHour;
	rtc_tm->tm_mday	= pTime.wDay;
	rtc_tm->tm_mon	= pTime.wMonth - 1;
	rtc_tm->tm_year	= pTime.wYear - 1900;

	pr_debug("read time %02d.%02d.%02d %02d/%02d/%02d\n",
		 rtc_tm->tm_year, rtc_tm->tm_mon, rtc_tm->tm_mday,
		 rtc_tm->tm_hour, rtc_tm->tm_min, rtc_tm->tm_sec);

    local_irq_enable();

    return 0;
}

static int tcc_rtc_settime(struct device *dev, struct rtc_time *tm)
{
    rtctime pTime;

	pr_debug("set time %02d.%02d.%02d %02d/%02d/%02d\n",
		 tm->tm_year, tm->tm_mon, tm->tm_mday,
		 tm->tm_hour, tm->tm_min, tm->tm_sec);

    local_irq_disable();
    pTime.wSecond       = tm->tm_sec;
    pTime.wMinute       = tm->tm_min;
    pTime.wHour         = tm->tm_hour;
    pTime.wDay          = tm->tm_mday;
    pTime.wMonth        = tm->tm_mon + 1;
    pTime.wYear         = tm->tm_year + 1900;

    tca_rtc_settime(&pTime);

    local_irq_enable();
	return 0;
}

static int tcc_rtc_getalarm(struct device *dev, struct rtc_wkalrm *alrm)
{
    rtctime pTime;
	struct rtc_time *alm_tm = &alrm->time;
	unsigned int alm_en;

    local_irq_disable();

    alm_en = HwRTCALM;
    alrm->enabled = (alm_en & HwRTCALM_ALMEN_EN) ? 1 : 0;
    alrm->pending = (alm_en & HwRTCPEND_PENDING) ? 1 : 0;

    pr_debug("tcc_rtc_getalarm : enabled = %d, pending = %d\n", alrm->enabled, alrm->pending);

    tca_alarm_gettime(&pTime);

	alm_tm->tm_sec  = pTime.wSecond;
	alm_tm->tm_min  = pTime.wMinute;
	alm_tm->tm_hour = pTime.wHour;
	alm_tm->tm_mday = pTime.wDay ;
	alm_tm->tm_mon  = pTime.wMonth - 1;
	alm_tm->tm_year = pTime.wYear - 1900;


	pr_debug("read alarm[%d] %02d.%02d.%02d %02d/%02d/%02d\n",
		 alm_en, alm_tm->tm_year, alm_tm->tm_mon, alm_tm->tm_mday,
		 alm_tm->tm_hour, alm_tm->tm_min, alm_tm->tm_sec);

    local_irq_enable();
	return 0;
}

static int tcc_rtc_setalarm(struct device *dev, struct rtc_wkalrm *alrm)
{
    rtctime pTime;


    struct rtc_time *tm = &alrm->time;

    local_irq_disable();

    alrm->enabled = 1;

	pr_debug("set alarm %02d.%02d.%02d %02d/%02d/%02d\n",
		 tm->tm_year, tm->tm_mon, tm->tm_mday,
		 tm->tm_hour, tm->tm_min, tm->tm_sec);

    pTime.wSecond   = tm->tm_sec;
    pTime.wMinute   = tm->tm_min;
    pTime.wHour     = tm->tm_hour;
    pTime.wDay      = tm->tm_mday;
    pTime.wMonth    = tm->tm_mon + 1;
    pTime.wYear     = tm->tm_year + 1900;

    tca_alarm_settime(&pTime);

    if(alrm->enabled){
        HwRTCALM |= HwRTCALM_ALMEN_EN;
        enable_irq_wake(tcc_rtc_alarmno);
    }else {
        HwRTCALM &= ~HwRTCALM_ALMEN_EN;
        disable_irq_wake(tcc_rtc_alarmno);
    }

    local_irq_enable();


	return 0;
}

static int tcc_rtc_ioctl(struct device *dev,
			 unsigned int cmd, unsigned long arg)
{
	unsigned int ret = -ENOIOCTLCMD;

	switch (cmd) {
	case RTC_AIE_OFF:
        break;
	case RTC_AIE_ON:
		tcc_rtc_setaie((cmd == RTC_AIE_ON) ? 1 : 0);
		ret = 0;
		break;
    case RTC_PIE_OFF:
        break;
    case RTC_PIE_ON:
        break;
    case RTC_IRQP_READ:
        break;
    case RTC_IRQP_SET:
        break;
    case RTC_UIE_ON:
        break;
    case RTC_UIE_OFF:
        break;
        ret = -EINVAL;
    }

	return ret;
}


static struct rtc_class_ops tcc_rtcops = {
  	.read_time	= tcc_rtc_gettime,
  	.set_time	= tcc_rtc_settime,  	
	.ioctl		= tcc_rtc_ioctl,
	.read_alarm	= tcc_rtc_getalarm,
	.set_alarm	= tcc_rtc_setalarm,
};


#define	IO_CKC_EnableBUS_RTC()		(BITSET(HwBCLKCTR, HwBCLKCTR_RTC_ON))

#if 0
static void tcc_rtc_enable(struct platform_device *pdev)
{
    IO_CKC_EnableBUS_RTC();
    HwRTCCON |= HwRTCCON_RTCEN;
    HwINTCON |= HwINTCON_INTWREN;    
    HwINTCON    = 0;
    HwRTCPEND   = 0;
    HwINTCON    &= ~HwINTCON_INTWREN;
    HwRTCCON    &= ~HwRTCCON_RTCEN;

}
#endif

static int tcc_rtc_remove(struct platform_device *dev)
{
	struct rtc_device *rtc = platform_get_drvdata(dev);

	platform_set_drvdata(dev, NULL);
	rtc_device_unregister(rtc);

    tcc_rtc_setaie(0);

    free_irq(tcc_rtc_alarmno, rtc);

	return 0;
}

static int tcc_rtc_probe(struct platform_device *pdev)
{
	struct rtc_device *rtc;
	int ret;

    pr_debug("%s: probe=%p\n", __func__, pdev);

    tca_alarm_setint();
    //tcc_rtc_enable(pdev);
	
    tcc_rtc_alarmno = platform_get_irq(pdev, 0);

    /* ADD : Error handling */

    pr_debug("tcc_rtc: alarm irq %d\n", tcc_rtc_alarmno);

	/* register RTC and exit */

	rtc = rtc_device_register(pdev->name, &pdev->dev, &tcc_rtcops, THIS_MODULE);

	if (IS_ERR(rtc)) {
		dev_err(&pdev->dev, "cannot attach rtc\n");
		ret = PTR_ERR(rtc);
		goto err_nores;
	}

    platform_set_drvdata(pdev, rtc);



    if(request_irq(tcc_rtc_alarmno, &tcc_rtc_alarmirq,
          IRQF_DISABLED,   DRV_NAME, rtc))
    {
        pr_debug("%s: RTC timer interrupt IRQ%d already claimed\n",
                 pdev->name, tcc_rtc_alarmno);
       goto fail;
    }

//    tca_alarm_setpmwkup();
//    device_init_wakeup(&pdev->dev, 0);

	return 0;

fail:
    return 0;

err_nores:
	rtc_device_unregister(rtc);
	return ret;
}

#ifdef CONFIG_PM

/* RTC Power management control */
static int tcc_rtc_suspend(struct platform_device *pdev, pm_message_t state)
{
    if (device_may_wakeup(&pdev->dev))
        enable_irq_wake(tcc_rtc_alarmno);

    return 0;
}

static int tcc_rtc_resume(struct platform_device *pdev)
{
    if (device_may_wakeup(&pdev->dev))
        disable_irq_wake(tcc_rtc_alarmno);

    return 0;
}
#else
#define tcc_rtc_suspend NULL
#define tcc_rtc_resume  NULL
#endif

static struct platform_driver tcc_rtcdrv = {
	.probe		= tcc_rtc_probe,
	.remove		= tcc_rtc_remove,
	.suspend	= tcc_rtc_suspend,
	.resume		= tcc_rtc_resume,
	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
	},
};

static char __initdata banner[] = "TCC RTC, (c) 2008, Telechips \n";

static int __init tcc_rtc_init(void)
{
	printk(banner);
	return platform_driver_register(&tcc_rtcdrv);
}

static void __exit tcc_rtc_exit(void)
{
	platform_driver_unregister(&tcc_rtcdrv);
}

module_init(tcc_rtc_init);
module_exit(tcc_rtc_exit);

MODULE_AUTHOR("linux <linux@telechips.com>");
MODULE_DESCRIPTION("Telechips RTC Driver");
MODULE_LICENSE("GPL");
