
/* linux/drivers/spi/spi_tcc83x.c */

#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/platform_device.h>

#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/hardware.h>

#include <asm/arch/tcc83x_virt.h>
#include <linux/workqueue.h>

#include <asm/plat-tcc/tcc_ckc.h>

#define TCC_MAX_FRAME_BYTE 256
#define TCC_TX_FIFO_CNT 8
#define TCC_TX_FIFO_UNIT_SIZE 4
#define TCC_TX_FIFO_THRESHOLD 4
#define TCC_TX_FIFO_WRITEABLE_CNT (TCC_TX_FIFO_CNT - TCC_TX_FIFO_THRESHOLD)
#define TCC_RX_FIFO_THRESHOLD 4

#define MAX_WAIT_TIME 5000
#define MILLISEC_TO_TICK(X) (((X) * HZ) / 1000)

#pragma pack(push, 4)
struct tcc83x_spi_regs {
    volatile unsigned long SDO, SDI, SBCR, SCR, SPFRMPOS, SPCTRL, SPSTS;
};
#pragma pack(pop)

struct tcc83x_spi {
    spinlock_t lock;
    int tx_irq, rx_irq;
    struct platform_device *pdev;
    volatile struct tcc83x_spi_regs *spi_regs;

    u8 stopping;
    struct list_head queue;
    struct spi_message *current_message;

    unsigned char rx[TCC_MAX_FRAME_BYTE];
    unsigned char tx[TCC_MAX_FRAME_BYTE];
    int rx_len, rx_cnt;
    int tx_len, tx_cnt;

    int rx_threshold;
    wait_queue_head_t wait_q;
    struct {
        int xfer_pos;
        int value;
    } done;
};

/* #set RCFul, ROR, RUR  #clear RCFul, REmp, RFul, ROR, RUR at SPCTRL */
#define tcc83x_en_rxint(tspi) BITSET((tspi)->spi_regs->SPCTRL, Hw12 | Hw9 | Hw8)
#define tcc83x_dis_rxint(tspi) BITCLR((tspi)->spi_regs->SPCTRL, Hw12 | Hw11 | Hw10 | Hw9 | Hw8)

/* #set TCEmpEn, TOREn, TUREn  #clear TCEmpEn, TEmpEn, TFulEn, TOREn, TUREn at SPCTRL */
#define tcc83x_en_txint(tspi) BITSET((tspi)->spi_regs->SPCTRL, Hw4 | Hw1 | Hw0)
#define tcc83x_dis_txint(tspi) BITCLR((tspi)->spi_regs->SPCTRL, Hw4 | Hw3 | Hw2 | Hw1 | Hw0)

/* #set EN  #clear EN at SCR */
#define tcc83x_en_spi(tspi) BITSET((tspi)->spi_regs->SCR, Hw31)
#define tcc83x_dis_spi(tspi) BITCLR((tspi)->spi_regs->SCR, Hw31)

static void tcc83x_set_rx_threshold(struct tcc83x_spi *tspi, unsigned int th)
{
    th &= 0x7;
    tspi->rx_threshold = th;
 
    /* clear RxCnt and set RxCnt */
    BITCLR(tspi->spi_regs->SPCTRL, Hw13 | Hw14 | Hw15);
    BITSET(tspi->spi_regs->SPCTRL, (th << 13));
}

static void tcc83x_set_tx_threshold(struct tcc83x_spi *tspi, unsigned int th)
{
    BITCLR(th, ~(Hw0 | Hw1 | Hw2));

    /* clear TxCnt and set TxCnt */
    BITCLR(tspi->spi_regs->SPCTRL, Hw5 | Hw6 | Hw7);
    BITSET(tspi->spi_regs->SPCTRL, (th << 5));
}

static void tcc83x_set_frame_length(struct tcc83x_spi *tspi, unsigned int len)
{
    len = (len - 1) & 0xFF;

    /* clear FRM_LEN and set FRM_LEN */
    BITCLR(tspi->spi_regs->SCR,  Hw16 | Hw17 | Hw18 | Hw19 | Hw20 | Hw21 | Hw22 | Hw23);
    BITSET(tspi->spi_regs->SCR, len << 16);
}

static void tcc83x_clear_fifos(struct tcc83x_spi *tspi)
{
    unsigned int reg = 0;
    /* #set TEmpEn  #clear TEmpEnat SPCTRL */
    BITSET(tspi->spi_regs->SPCTRL, Hw3);
    tcc83x_en_spi(tspi);
    while (ISCLR(tspi->spi_regs->SPSTS, Hw3)) {
        schedule();
    }
    BITCLR(tspi->spi_regs->SPCTRL, Hw3);

    udelay(2000);
    udelay(1200);

    tcc83x_dis_spi(tspi);
    /* #set REmpEn  #clear REmpEnat SPCTRL */
    BITSET(tspi->spi_regs->SPCTRL, Hw11);
    while (ISCLR(tspi->spi_regs->SPSTS, Hw11)) {
        reg = HwSDI;
        schedule();
    }
    BITCLR(tspi->spi_regs->SPCTRL, Hw11);
}

static int tcc83x_spi_start_rxtx(struct spi_master *master, unsigned int flen)
{
    int ret = -EIO;
    struct tcc83x_spi *tspi = spi_master_get_devdata(master);

    if ((flen > 0) && (flen <= TCC_MAX_FRAME_BYTE)) {
        tspi->rx_len = flen;
        tspi->tx_len = flen;
        tspi->rx_cnt = 0;
        tspi->tx_cnt = 0;

        if (flen & 0xF) { tcc83x_set_rx_threshold(tspi, 1); }

        tcc83x_set_frame_length(tspi, flen);
        tcc83x_en_rxint(tspi);
        tcc83x_en_txint(tspi);
        tcc83x_en_spi(tspi);

        if (interruptible_sleep_on_timeout(&(tspi->wait_q),
                                           MILLISEC_TO_TICK(MAX_WAIT_TIME)) == 0) {
            printk("[%s:%d] sleep_on(%dms) timeout !!!",
                   __FUNCTION__, __LINE__, MAX_WAIT_TIME);
        }
        if (tspi->rx_cnt == tspi->rx_len) {
            ret = 0;
        }

        if (flen & 0xF) { tcc83x_set_rx_threshold(tspi, TCC_RX_FIFO_THRESHOLD); }
    }

    tspi->current_message = NULL;
    tcc83x_dis_spi(tspi);
    tcc83x_dis_rxint(tspi);
    tcc83x_dis_txint(tspi);
    return ret;
}

static void tcc83x_spi_next_message(struct spi_master *master)
{
    struct tcc83x_spi *tspi = NULL;
    struct spi_message *msg = NULL;
    struct spi_device *spi = NULL;
    struct spi_transfer *xfer = NULL;
    unsigned int flen = 0;
    unsigned int copy_len = 0, total_len = 0;

    tspi = spi_master_get_devdata(master);

    BUG_ON(tspi->current_message);

    msg = list_entry(tspi->queue.next, struct spi_message, queue);
    list_del_init(&(msg->queue));
    tspi->current_message = msg;

    spi = msg->spi;

    list_for_each_entry(xfer, &(msg->transfers), transfer_list) {
        tspi->done.xfer_pos = 0;
        tspi->done.value = 0;
        total_len = xfer->len;
        do {
            copy_len = (total_len > TCC_MAX_FRAME_BYTE) ? TCC_MAX_FRAME_BYTE : total_len;
            if (xfer->tx_buf) {
                memcpy(tspi->tx, xfer->tx_buf + tspi->done.xfer_pos, copy_len);
            }

            if ((tcc83x_spi_start_rxtx(master, copy_len))) {
                tcc83x_clear_fifos(tspi);
                goto lb_return;
            }

            if (xfer->rx_buf)
                memcpy(xfer->rx_buf + tspi->done.xfer_pos, tspi->rx, copy_len);

            total_len -= copy_len;
            tspi->done.xfer_pos += copy_len;
        } while (total_len);
        flen += xfer->len;
    }

lb_return:
    msg->status = tspi->done.value;
    if (msg->status == 0) {
        msg->actual_length = flen;
    }
    msg->complete(msg->context);

    if (list_empty(&(tspi->queue)) || tspi->stopping) {
        ;
    } else {
        tcc83x_spi_next_message(master);
    }
}

static irqreturn_t tcc83x_spi_irq_tx(int irq, void *dev_id)
{
    struct spi_master *master = dev_id;
    struct tcc83x_spi *tspi = spi_master_get_devdata(master);
    struct spi_message *msg = NULL;
    unsigned int reg_temp = 0, dsize = 0, data = 0;
    unsigned long flags = 0;
    int i = 0, done_value = -1;

    local_save_flags(flags);
    local_irq_disable();

    reg_temp = tspi->spi_regs->SPSTS;
    msg = tspi->current_message;

    if (ISSET(reg_temp, Hw3)) {
        BITCLR(tspi->spi_regs->SPCTRL, Hw3);
    }

    if (reg_temp & Hw4) {
        for (i = 0; i < TCC_TX_FIFO_WRITEABLE_CNT; i++) {
            if (tspi->tx_cnt < tspi->tx_len) {
                dsize = tspi->tx_len - tspi->tx_cnt;
                if (dsize > 4) { dsize = 4; }
                memcpy(&data, tspi->tx + tspi->tx_cnt, dsize);
                tspi->spi_regs->SDO = data;
                tspi->tx_cnt += dsize;
            } else {
                tcc83x_dis_txint(tspi);
                break;
            }
        }
    }

    if (reg_temp & (Hw1 | Hw0)) {
        printk("%s : %s "
               "(rx_cnt[%d], tx_cnt[%d]), (rx_len[%d], tx_len[%d])\n",
               __func__, (reg_temp & Hw1) ? "overrun" : "underrun",
               tspi->rx_cnt, tspi->tx_cnt, tspi->rx_len, tspi->tx_len);
        tspi->done.value = done_value = -EIO;
    }

    if (done_value != -1) {
        wake_up(&(tspi->wait_q));
    }


    (irq == INT_SPDTX) ? BITSET(HwCLR, Hw9) : BITSET(HwCLR_B, Hw10);
    local_irq_restore(flags);
    return IRQ_HANDLED;
}

static irqreturn_t tcc83x_spi_irq_rx(int irq, void *dev_id)
{
    struct spi_master *master = dev_id;
    struct tcc83x_spi *tspi = spi_master_get_devdata(master);
    struct spi_message *msg = NULL;
    unsigned int reg_temp = 0, dsize = 0, data = 0;
    unsigned long flags = 0;
    int i = 0, done_value = -1;

    local_save_flags(flags);
    local_irq_disable();

    reg_temp = tspi->spi_regs->SPSTS;
    msg = tspi->current_message;

    if (ISSET(reg_temp, Hw11)) {
        BITCLR(tspi->spi_regs->SPCTRL, Hw11);
    }

    if (reg_temp & Hw12) {
        for (i = 0; i < tspi->rx_threshold; i++) {
            if (tspi->rx_cnt < tspi->rx_len) {
                data = tspi->spi_regs->SDI;
                dsize = tspi->rx_len - tspi->rx_cnt;
                if (dsize > 4) { dsize = 4; }
                memcpy(tspi->rx + tspi->rx_cnt, &data, dsize);
                tspi->rx_cnt += dsize;
            }
        }

        if (tspi->rx_cnt >= tspi->rx_len) {
            tcc83x_dis_rxint(tspi);
            done_value = 0;
        }
    }

    if (reg_temp & (Hw9 | Hw8)) {
        printk("%s : %s "
               "(rx_cnt[%d], tx_cnt[%d]), (rx_len[%d], tx_len[%d])\n",
               __func__, (reg_temp & Hw9) ? "overrun" : "underrun",
               tspi->rx_cnt, tspi->tx_cnt, tspi->rx_len, tspi->tx_len);
        tspi->done.value = done_value = -EIO;
    }

    if (done_value != -1) {
        wake_up(&(tspi->wait_q));
    }

    (irq == INT_SPI_R) ? BITSET(HwCLR, Hw28) : BITSET(HwCLR_B, Hw11);
    local_irq_restore(flags);
    return IRQ_HANDLED;
}

#define MODEBITS (SPI_CPOL | SPI_CPHA | SPI_CS_HIGH | SPI_LSB_FIRST)
static int tcc83x_spi_setup(struct spi_device *spi)
{
    struct tcc83x_spi *tspi = NULL;
    unsigned int bits = spi->bits_per_word;
    unsigned int scr = 0;

    tspi = spi_master_get_devdata(spi->master);

    if (tspi->stopping)
        return -ESHUTDOWN;
    if (spi->chip_select > spi->master->num_chipselect) {
        dev_dbg(&spi->dev,
                "setup: invalid chipselect %u (%u defined)\n",
                spi->chip_select, spi->master->num_chipselect);
        return -EINVAL;
    }

    if (bits == 0)
        bits = 8;
    if (bits != 8 && bits != 16 && bits != 32) {
        dev_dbg(&spi->dev, "setup: invalid bits_per_word %u\n", bits);
        return -EINVAL;
    }

    if (spi->mode & ~MODEBITS) {
        dev_dbg(&spi->dev, "setup: unsupported mode bits %x\n",
                spi->mode & ~MODEBITS);
        return -EINVAL;
    }

    scr = tspi->spi_regs->SCR;
    scr = (spi->mode & SPI_CPOL) ? (scr & ~Hw29) : (scr | Hw29);
    scr = (spi->mode & SPI_CPHA) ? (scr | Hw28) : (scr & ~Hw28);
    scr = (spi->mode & SPI_CS_HIGH) ? (scr | Hw12) : (scr & ~Hw12);
    scr = (spi->mode & SPI_LSB_FIRST) ? (scr & ~Hw27) : (scr | Hw27);
    tspi->spi_regs->SCR = scr;

    return 0;
}

static int tcc83x_spi_transfer(struct spi_device *spi, struct spi_message *msg)
{
    struct tcc83x_spi *tspi = NULL;
    unsigned long flags = 0;

    tspi = (struct tcc83x_spi *)spi_master_get_devdata(spi->master);

    if (unlikely(list_empty(&(msg->transfers)))) {
        return -EINVAL;
    }

    if (tspi->stopping) {
        return -ESHUTDOWN;
    }

    msg->status = -EINPROGRESS;
    msg->actual_length = 0;

    spin_lock_irqsave(&(tspi->lock), flags);
    list_add_tail(&(msg->queue), &(tspi->queue));
    if (!tspi->current_message) {
        tcc83x_spi_next_message(spi->master);
    }
    spin_unlock_irqrestore(&(tspi->lock), flags);
    return 0;
}

static void tcc83x_spi_cleanup(struct spi_device *spi)
{
    if (!spi->controller_state)
        return;
    dev_dbg(&spi->dev, "tcc8300_spi_cleanup\n");
}

static int __init tcc83x_spi_probe(struct platform_device *pdev)
{
    struct resource *regs = NULL;
    int tx_irq = 0, rx_irq = 0, ret = 0;
    struct spi_master *master = NULL;
    struct tcc83x_spi *tspi = NULL;

    printk("[TCC SPI] *** - Interrupt mode \n");
    regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (!regs) {
        return -ENXIO;
    }
    tx_irq = platform_get_irq(pdev, 0);
    rx_irq = platform_get_irq(pdev, 1);
    if (tx_irq < 0 || rx_irq < 0) {
        return -ENXIO;
    }

    master = spi_alloc_master(&(pdev->dev), sizeof(struct tcc83x_spi));
    if (!master) {
        ret = -ENOMEM;
        goto out_free;
    }

    master->bus_num = pdev->id;
    master->num_chipselect = 1;
    master->setup = tcc83x_spi_setup;
    master->transfer = tcc83x_spi_transfer;
    master->cleanup = tcc83x_spi_cleanup;
    platform_set_drvdata(pdev, master);

    tspi = spi_master_get_devdata(master);

    spin_lock_init(&(tspi->lock));
    INIT_LIST_HEAD(&(tspi->queue));

    tspi->pdev = pdev;
    tspi->spi_regs = (volatile struct tcc83x_spi_regs *)regs->start;
    tspi->tx_irq = tx_irq;
    tspi->rx_irq = rx_irq;
    if (tspi->spi_regs == 0) {
        goto out_free_buffer;
    }
    init_waitqueue_head(&(tspi->wait_q));

    if (pdev->id == 0) {
        BITSET(HwBCLKCTR, Hw24); /* SPI0 Master0 Clock Enable */
        BITSET(HwSWRESET, Hw24); /* SPI0 Master0 Reset Enable */
        BITCLR(HwSWRESET, Hw24); /* SPI0 Master0 Reset Disable */

        BITCLR(HwNGSEL_B, Hw3 | Hw2 | Hw1 | Hw0); /* GPIO_B[3:0] is not general I/O */
        BITCLR(HwGSEL_B, Hw10); /* GPIO_B[3:0] is not use UART0 */
        BITSET(HwGSEL_B, Hw14); /* GPIO_B[3:0] is SPI0 */

        BITCLR(HwIEN, Hw9 | Hw28);
        BITSET(HwSEL, Hw9 | Hw28);
        BITSET(HwMODE, Hw9 | Hw28);
        BITCLR(HwPOL, Hw9 | Hw28);
        BITSET(HwIEN, Hw9 | Hw28);
        tca_ckc_setperi(PERI_SPIMS0, KHz, 2400, DIRECTPLL0);
    } else {
        BITSET(HwBCLKCTR, Hw30); /* SPI1 Master0 Clock Enable */
        BITSET(HwSWRESET, Hw30); /* SPI1 Master0 Reset Enable */
        BITCLR(HwSWRESET, Hw30); /* SPI1 Master0 Reset Disable */

        BITCLR(HwNGSEL_F, (Hw3 | Hw2 | Hw1 | Hw0)); /* GPIO_F[3:0] is SPI1 */
        BITCLR(HwGSEL_F, Hw10); /* GPIO_F[3:0] is not use UART2 */
        BITSET(HwGSEL_F, Hw14); /* GPIO_F[3:0] is SPI1 */

        BITCLR(HwIEN_B, Hw10 | Hw11);
        BITSET(HwSEL_B, Hw10 | Hw11);
        BITSET(HwMODE_B, Hw10 | Hw11);
        BITCLR(HwPOL_B, Hw10 | Hw11);
        BITSET(HwIEN_B, Hw10 | Hw11);
        tca_ckc_setperi(PERI_SPIMS1, KHz, 2400, DIRECTPLL0);
    }

    tspi->spi_regs->SBCR = 0;
    tcc83x_set_rx_threshold(tspi, TCC_RX_FIFO_THRESHOLD);
    tcc83x_set_tx_threshold(tspi, TCC_TX_FIFO_THRESHOLD);
    tspi->spi_regs->SCR = (Hw30 | /* Master */
                           Hw29 | /* SCK Polarity */
                           Hw27 | /* MSB/LSB */
                           Hw26 | /* TDO Enable */
                           (TCC_MAX_FRAME_BYTE - 1) << 16 | /* frame length */
                           Hw14 /* Tx Mode Enable */);

    BITSET((tspi)->spi_regs->SCR, Hw10); /* Clear RX FIFO */
    BITSET((tspi)->spi_regs->SCR, Hw9);  /* Clear TX FIFO */
    BITSET((tspi)->spi_regs->SCR, Hw26); /* Enable TX Data Out */

    ret = request_irq(tx_irq, tcc83x_spi_irq_tx, IRQF_SHARED, pdev->dev.bus_id, master);
    if (ret) { goto out_unmap_regs1; }

    ret = request_irq(rx_irq, tcc83x_spi_irq_rx, IRQF_SHARED, pdev->dev.bus_id, master);
    if (ret) { goto out_unmap_regs2; }

    tcc83x_clear_fifos(tspi);
    tcc83x_dis_spi(tspi);

    ret = spi_register_master(master);
    if (ret) { goto out_reset_hw; }

    return 0;

out_reset_hw:
    free_irq(rx_irq, master);
out_unmap_regs2:
    free_irq(tx_irq, master);
out_unmap_regs1:
out_free_buffer:
out_free:
    spi_master_put(master);
    return ret;
}

static int __exit tcc83x_spi_remove(struct platform_device *dev)
{
    struct spi_master *master = platform_get_drvdata(dev);
    struct tcc83x_spi *tspi = NULL;
    struct spi_message *msg = NULL;

    tspi = spi_master_get_devdata(master);

    spin_lock_irq(&(tspi->lock));
    tspi->stopping = 1;
    spin_unlock_irq(&tspi->lock);

    list_for_each_entry(msg, &(tspi->queue), queue) {
        msg->status = -ESHUTDOWN;
        msg->complete(msg->context);
    }

    free_irq(tspi->rx_irq, master);
    free_irq(tspi->tx_irq, master);

    spi_unregister_master(master);
    return 0;
}

static struct platform_driver tcc83x_spidrv = {
    .remove = __exit_p(tcc83x_spi_remove),
    .driver = {
        .name = "tcc83x-spi",
        .owner = THIS_MODULE,
    },
};

static int __init tcc83x_spi_init(void)
{
    return platform_driver_probe(&tcc83x_spidrv, tcc83x_spi_probe);
}

static void __exit tcc83x_spi_exit(void)
{
    platform_driver_unregister(&tcc83x_spidrv);
}

module_init(tcc83x_spi_init);
module_exit(tcc83x_spi_exit);

MODULE_DESCRIPTION("TCC83X SPI Driver");
MODULE_LICENSE("GPL");

