/*
 * linux/drivers/video/tcc92xx_blit.c
 *
 * Copyright (C) 2008-2009 Telechips 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/workqueue.h>
#include <linux/wait.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/file.h>

#include <asm/tlbflush.h>
#include <asm/cacheflush.h>


#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/div64.h>
#include <asm/mach/map.h>
#ifdef CONFIG_PM
#include <linux/pm.h>
#endif
#include <linux/major.h>

#include <mach/bsp.h>
#include <mach/tca_ckc.h>

#ifdef CONFIG_ANDROID_PMEM
#include <linux/android_pmem.h>
#endif

#include "mach/tcc92xxfb_ioctrl.h"
#include "mach/tcc_grp_ioctrl.h"
#include "mach/tcc_scaler_ioctrl.h"

static int debug = 0;
#define dprintk(msg...)	if (debug) { printk( "tcc_blit: " msg); }

static struct clk *overlay_clk;
static struct clk *scaler_clk;

DEFINE_MUTEX(blit_mutex);

void grp_copyblit_ctrl(G2D_BITBLIT_TYPE *g2d_p);
char tcc_g2d_working_check(void);

void tcc_scaler_init(void);
char M2M_Scaler1_Ctrl_Detail(SCALER_TYPE *scale_img);

int get_img(struct tcc_img *img, struct fb_info *info,
	    unsigned long *start, unsigned long *len,
	    struct file** filep)
{
	int put_needed, ret = 0;
	struct file *file;
	unsigned long vstart;

#ifdef CONFIG_ANDROID_PMEM
	if (!get_pmem_file(img->fd, start, &vstart, len, filep))
		return 0;
#endif

	file = fget_light(img->fd, &put_needed);
	if (file == NULL)
	{
		printk("%s fget_light \n", __func__);
		return -1;
	}

	if (MAJOR(file->f_dentry->d_inode->i_rdev) == FB_MAJOR) {
		*start = info->fix.smem_start;
		*len = info->fix.smem_len;
	} else	{
		printk("%s MAJOR error \n", __func__);
		ret = -1;
	}

	
	fput_light(file, put_needed);

	return ret;
}

void put_img(struct file *src_file, struct file *dst_file)
{
#ifdef CONFIG_ANDROID_PMEM
	if (src_file)
		put_pmem_file(src_file);
	if (dst_file)
		put_pmem_file(dst_file);
#endif
}

SCALER_DATA_FM TCC_ChFmt_scaler(G2D_DATA_FM g2d_fmt)
{
	SCALER_DATA_FM fmt = GE_RGB565;

	switch(g2d_fmt)
	{
		case GE_YUV444_sp:
			break;

	    case GE_YUV440_sp:
			break;

	    case GE_YUV422_sp:
			fmt = SCALER_YUV422_sp;
			break;

	    case GE_YUV420_sp:
			fmt = SCALER_YUV420_sp;
			break;

	    case GE_YUV411_sp:
			break;

	    case GE_YUV410_sp:
			break;

	    case GE_YUV422_in:
			break;

	    case GE_YUV420_in:
			break;

	    case GE_YUV444_sq:
			break;

	    case GE_YUV422_sq:
			break;

		case GE_RGB332:
			break;

		case GE_RGB444:
			break;

	    case GE_ARGB4444:
			break;

	    case GE_RGB454:
			break;

	    case GE_ARGB3454:
			break;

	    case GE_RGB555:
			fmt = SCALER_RGB565;
			break;

	    case GE_ARGB1555:
			fmt = SCALER_RGB565;
			break;

	    case GE_RGB565:
			fmt = SCALER_RGB565;
			break;

	    case GE_RGB666:
			break;

	    case GE_ARGB4666:
			break;

	    case GE_ARGB6666:
			break;

	    case GE_RGB888:
			break;

	    case GE_ARGB4888:
			break;

	    case GE_ARGB8888:

			break;
	}

	return fmt;
}




int tcc92xx_blit(struct fb_info *fb, struct tcc_blit_req *req)
{
	unsigned long src_start = 0, src_len = 0, dst_start = 0, dst_len = 0;
	struct file *src_file = 0;
	struct file *dst_file = 0;
	G2D_BITBLIT_TYPE g2d_p;
	SCALER_TYPE scale_img;
	int result = 0, scale_opt = 0;
	
	memset(&g2d_p, 0x00, sizeof(G2D_BITBLIT_TYPE));
	memset(&scale_img, 0x00, sizeof(SCALER_TYPE));


	if (unlikely(get_img(&req->src, fb, &src_start, &src_len, &src_file))) {
		dprintk(KERN_ERR "could not retrieve src image\n");
		return -EINVAL;
	}
	if (unlikely(get_img(&req->dst, fb, &dst_start, &dst_len, &dst_file))) {
		dprintk(KERN_ERR "could not retrieve dst image\n");
#ifdef CONFIG_ANDROID_PMEM
		put_pmem_file(src_file);
#endif
		return -EINVAL;
	}

	mutex_lock(&blit_mutex);

	local_flush_tlb_all();
	flush_cache_all();

	if(((req->transform == 0) && ((req->src_rect.w != req->dst_rect.w) || (req->src_rect.h != req->dst_rect.h)))
		||((req->transform != 0) && ((req->src_rect.w != req->dst_rect.h) || (req->src_rect.h != req->dst_rect.w))))
	{				
		#define SCALE_SIZE	8
		
		#if 0
		scale_opt += req->src.width % SCALE_SIZE;
		scale_opt += req->dst.width % SCALE_SIZE;

		scale_opt += req->src.height % SCALE_SIZE;
		scale_opt += req->dst.height % SCALE_SIZE;
		
		scale_opt += req->src_rect.x % SCALE_SIZE;
		scale_opt += req->src_rect.w % SCALE_SIZE;
		scale_opt += req->src_rect.y % SCALE_SIZE;
		scale_opt += req->src_rect.h % SCALE_SIZE;

		scale_opt += req->dst_rect.x % SCALE_SIZE;
		scale_opt += req->dst_rect.w % SCALE_SIZE;
		scale_opt += req->dst_rect.y % SCALE_SIZE;
		scale_opt += req->dst_rect.h % SCALE_SIZE;
		
		scale_opt += ((src_start + req->src.offset)&0x7);
		scale_opt += ((dst_start + req->dst.offset)&0x7);

		if(scale_opt == 0)
		{
			scale_img.responsetype = SCALER_INTERRUPT;
			scale_img.src_Yaddr = src_start + req->src.offset; 	// source address
			scale_img.src_Uaddr = 0; 	// source address
			scale_img.src_Vaddr = 0; 	// source address
			scale_img.src_fmt = TCC_ChFmt_scaler(req->src.format);		// source image format
			scale_img.src_ImgWidth = req->src.width;	// source image width
			scale_img.src_ImgHeight = req->src.height;	// source image height

			scale_img.src_winLeft = req->src_rect.x;
			scale_img.src_winTop = req->src_rect.y;
			scale_img.src_winRight = req->src_rect.x + req->src_rect.w;
			scale_img.src_winBottom = req->src_rect.y + req->src_rect.h;

			scale_img.dest_Yaddr = dst_start + req->dst.offset;	// destination image address
			scale_img.dest_Uaddr = 0;	// destination image address
			scale_img.dest_Vaddr = 0;	// destination image address
			scale_img.dest_fmt = TCC_ChFmt_scaler(req->dst.format);		// destination image format
			scale_img.dest_ImgWidth = req->dst.width;	// destination image width
			scale_img.dest_ImgHeight = req->dst.height; // destination image height

			scale_img.dest_winLeft = req->dst_rect.x;
			scale_img.dest_winTop = req->dst_rect.y;
			scale_img.dest_winRight =req->dst_rect.x + req->dst_rect.h;
			scale_img.dest_winBottom = req->dst_rect.y + req->dst_rect.w;

			clk_enable(scaler_clk);
			grp_copyblit_ctrl(&g2d_p);
			clk_disable(scaler_clk);

			M2M_Scaler1_Ctrl_Detail(&scale_img);

			result = -EINVAL;

				
		}
		else
		#endif//			
		{
			dprintk("scaler fale :src_start=0x%lx, dst_start=0x%lx\n", (src_start + req->src.offset), (dst_start + req->dst.offset));
			dprintk("scaler fale : W %d %d H: %d %d \n", req->src.width , req->dst.width, req->src.height, req->dst.height);
			result = -EINVAL;
		}

	}
	else
	{	
		if (unlikely(req->src_rect.x > req->src.width ||
				 req->src_rect.y > req->src.height)) {
			dprintk(KERN_ERR "tcc_blit: src img rect is outside of img\n");
		}
		if (unlikely(req->dst_rect.x > req->dst.width ||
				 req->dst_rect.y > req->dst.height)) {
			dprintk(KERN_ERR "tcc_blit: dst img rect is outside of img\n");
		}

		if(!tcc_g2d_working_check())
		{
			g2d_p.responsetype = G2D_INTERRUPT;

			if(req->src.addr != 0)
				g2d_p.src0 = (unsigned int)req->src.addr;
			else
				g2d_p.src0 = src_start + req->src.offset;
				
			g2d_p.srcfm = req->src.format;
			g2d_p.src_imgx = req->src.width;
			g2d_p.src_imgy = req->src.height;
			g2d_p.ch_mode = req->transform;

			g2d_p.tgt0 = dst_start;
			g2d_p.tgtfm = req->dst.format;
			g2d_p.crop_offx = req->src_rect.x;
			g2d_p.crop_offy = req->src_rect.y;
			g2d_p.crop_imgx = req->src_rect.w;
			g2d_p.crop_imgy =  req->src_rect.h;

			g2d_p.tgt0 = dst_start + req->dst.offset;
			g2d_p.dst_imgx = req->dst.width;
			g2d_p.dst_imgy = req->dst.height;
			g2d_p.dst_off_x = req->dst_rect.x; //???
			g2d_p.dst_off_y = req->dst_rect.y; //???

			if((req->src.format == GE_ARGB8888)||(req->src.format == GE_ARGB4444))
				g2d_p.alpha_type = G2d_ALAPH_RGB; //G2d_ALAPH_RGB
			else
				g2d_p.alpha_type = G2d_ALPHA_VALUE; //G2d_ALAPH_RGB

			g2d_p.alpha_value = req->alpha;
			
			clk_enable(overlay_clk);
			grp_copyblit_ctrl(&g2d_p);
			clk_disable(overlay_clk);
			
		}
		else
		{
			result = -EINVAL;
		}
	}

	mutex_unlock(&blit_mutex);
	
	return result;
}

static int __init tcc92xx_blit_init(void)
{

	overlay_clk = clk_get(0, "overlay");
	scaler_clk = clk_get(0, "m2m0");
//	tcc_scaler_init();

	return 0;
}

late_initcall(tcc92xx_blit_init);
