/*
 * linux/sound/soc/codecs/wm8524.c
 *
 * Based on:  wm8753.c by Liam Girdwood
 * Author: Richard Purdie <richard@openedhand.com>
 * Rewritten by:    <linux@telechips.com>
 * Created: June 10, 2008
 * Description: WM8524 ALSA SoC Audio driver
 *
 * Copyright 2005 Openedhand Ltd.
 * Copyright (C) 2008-2009 Telechips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/i2c.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <linux/gpio.h>

#include "wm8524.h"
/* Telechips */
#include "../tcc/tcc-pcm.h"
#include "../tcc/tca_tcchwcontrol.h"

#define WM8524_VERSION "0.01"

//#define alsa_dbg(f, a...)  printk("== alsa-debug == " f, ##a)
#define alsa_dbg(f, a...)

struct snd_soc_codec_device soc_codec_dev_wm8524;

/* codec private data */
struct wm8524_priv {
	unsigned int sysclk;
};

static int wm8524_hw_params(struct snd_pcm_substream *substream,
                            struct snd_pcm_hw_params *params, 
                            struct snd_soc_dai *dai)
{
    return 0;
}

static int wm8524_pcm_prepare(struct snd_pcm_substream *substream,
			      struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;

	/* set active */
#if defined(CONFIG_MACH_TCC9300ST)
    alsa_dbg("%s: \n", __func__);
    gpio_set_value(TCC_GPF(2), 1);      // Standby --> Enable mode
#endif
	return 0;
}

static void wm8524_shutdown(struct snd_pcm_substream *substream,
			    struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;

	/* deactivate */
	if (!codec->active) {
		udelay(50);
	}
}

static int wm8524_mute(struct snd_soc_dai *dai, int mute)
{
	struct snd_soc_codec *codec = dai->codec;

#if defined(CONFIG_MACH_TCC9300ST)
    alsa_dbg("%s: mute[%d]\n", __func__, mute);
	if (mute)
	{
		gpio_set_value(TCC_GPF(3), 1);      // MUTE
	}
	else
	{
		gpio_set_value(TCC_GPF(3), 0);      // Un-MUTE
	}
#endif

	return 0;
}

static int wm8524_set_dai_sysclk(struct snd_soc_dai *codec_dai,
                int clk_id, unsigned int freq, int dir)
{
    return 0;
}

static int wm8524_set_dai_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt)
{
    return 0;
}

static void wm8524_hw_free(struct snd_pcm_substream *substream)
{
#if defined(CONFIG_MACH_TCC9300ST)
    alsa_dbg("%s \n", __func__);
//    gpio_set_value(TCC_GPF(2), 0);      // Enable --> Standby mode
#endif
}


#define WM8524_RATES (SNDRV_PCM_RATE_44100)
#define WM8524_FORMATS (SNDRV_PCM_FMTBIT_S16_LE)

struct snd_soc_dai wm8524_dai = {
	.name = "WM8524",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 1,
		.channels_max = 2,
		.rates = WM8524_RATES,
		.formats = WM8524_FORMATS,},
	.ops = {
		.prepare = wm8524_pcm_prepare,
		.hw_params = wm8524_hw_params,
		.shutdown = wm8524_shutdown,
		.digital_mute = wm8524_mute,
		.set_sysclk = wm8524_set_dai_sysclk,
		.set_fmt = wm8524_set_dai_fmt,
		.hw_free = wm8524_hw_free,
	}
};
EXPORT_SYMBOL_GPL(wm8524_dai);

static int wm8524_suspend(struct platform_device *pdev, pm_message_t state)
{
#ifdef _WM8524_REG_DEBUG_	/* Telechips */
	printk("%s(): in...\n", __func__);
#endif

	return 0;
}

static int wm8524_resume(struct platform_device *pdev)
{
#ifdef _WM8524_REG_DEBUG_	/* Telechips */
	printk("%s(): out..\n", __func__);
#endif
	return 0;
}

static int wm8524_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct wm8524_setup_data *setup;
	struct snd_soc_codec *codec;
	struct wm8524_priv *wm8524;
	int ret = 0;

	pr_info("WM8524 Simple DAC %s", WM8524_VERSION);

	setup = socdev->codec_data;
	codec = kzalloc(sizeof(struct snd_soc_codec), GFP_KERNEL);
	if (codec == NULL)
		return -ENOMEM;

	wm8524 = kzalloc(sizeof(struct wm8524_priv), GFP_KERNEL);
	if (wm8524 == NULL) {
		kfree(codec);
		return -ENOMEM;
	}

	codec->name = "WM8524";
	codec->owner = THIS_MODULE;
	codec->dai = &wm8524_dai;
	codec->num_dai = 1;
	codec->write = NULL;
	codec->read = NULL;
	codec->private_data = wm8524;
		
	socdev->codec = codec;
	mutex_init(&codec->mutex);
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	/* register pcms */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0) {
		printk(KERN_ERR "wm8524: failed to create pcms\n");
		goto pcm_err;
	}

	ret = snd_soc_init_card(socdev);
	if (ret < 0) {
		printk(KERN_ERR "wm8524: failed to register card\n");
		goto card_err;
	}

	return ret;

card_err:
	snd_soc_free_pcms(socdev);
pcm_err:
	kfree(codec->private_data);
	kfree(codec);
	
	return ret;	
}

/* power down chip */
static int wm8524_remove(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

	snd_soc_free_pcms(socdev);
	kfree(codec->private_data);
	kfree(codec);

	return 0;
}

struct snd_soc_codec_device soc_codec_dev_wm8524 = {
	.probe = 	wm8524_probe,
	.remove = 	wm8524_remove,
	.suspend = 	wm8524_suspend,
	.resume =	wm8524_resume,
};
EXPORT_SYMBOL_GPL(soc_codec_dev_wm8524);

static int __init wm8524_modinit(void)
{
	return snd_soc_register_dai(&wm8524_dai);
}
module_init(wm8524_modinit);

static void __exit wm8524_exit(void)
{
	snd_soc_unregister_dai(&wm8524_dai);
}
module_exit(wm8524_exit);

MODULE_DESCRIPTION("ASoC WM8524 driver");
MODULE_AUTHOR("Richard Purdie");
MODULE_LICENSE("GPL");
