/*
 * linux/sound/soc/tcc/tcc9200-pcm.c  
 *
 * Based on:    linux/sound/arm/pxa2xx-pcm.c
 * Author:  <linux@telechips.com>
 * Created: Nov 30, 2004
 * Modified:    Nov 25, 2008
 * Description: ALSA PCM interface for the Intel PXA2xx chip
 *
 * Copyright:	MontaVista Software, Inc.
 * Copyright (C) 2008-2009 Telechips 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>

#include <linux/delay.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>

#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/bsp.h>
#include <mach/tca_ckc.h>

#include "tcc-i2s.h"
#include "tcc-pcm.h"

#include "tcc/tca_tcchwcontrol.h"

#define alsa_dbg(f, a...)  printk("== alsa-debug == " f, ##a)
//#define alsa_dbg(f, a...)

tcc_interrupt_info_x tcc_alsa_info;
EXPORT_SYMBOL(tcc_alsa_info);

static const struct snd_pcm_hardware tcc_pcm_hardware_play = {
    .info = (SNDRV_PCM_INFO_MMAP
             | SNDRV_PCM_INFO_MMAP_VALID
             | SNDRV_PCM_INFO_INTERLEAVED
             | SNDRV_PCM_INFO_BLOCK_TRANSFER
             | SNDRV_PCM_INFO_PAUSE
             | SNDRV_PCM_INFO_RESUME),

    .formats = (SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S24_LE),
    .rates      = SNDRV_PCM_RATE_8000_48000,
    .rate_min   = 8000,
    .rate_max   = 48000, 
    .channels_min = 2,
    .channels_max = 2,                                

#if 1
    .period_bytes_min = 16,
    .period_bytes_max = __play_buf_size,
    .periods_min = __play_buf_cnt,
    .periods_max = __play_buf_cnt,
    .buffer_bytes_max = __play_buf_cnt * __play_buf_size,
    .fifo_size = 16,  //ignored
#else
    .period_bytes_min = 128,
    .period_bytes_max = __play_buf_cnt * __play_buf_size,
    .periods_min = 1,
    .periods_max = __play_buf_cnt,
    .buffer_bytes_max = __play_buf_cnt * __play_buf_size,
    .fifo_size = 16,  //ignored
#endif
};

static const struct snd_pcm_hardware tcc_pcm_hardware_capture = {
    .info = (SNDRV_PCM_INFO_MMAP
             | SNDRV_PCM_INFO_MMAP_VALID
             | SNDRV_PCM_INFO_INTERLEAVED
             | SNDRV_PCM_INFO_BLOCK_TRANSFER
             | SNDRV_PCM_INFO_PAUSE
             | SNDRV_PCM_INFO_RESUME),

    .formats = SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S24_LE,
    .rates      = SNDRV_PCM_RATE_8000_48000,
    .rate_min   = 8000,
    .rate_max   = 48000, 
    .channels_min = 2,
    .channels_max = 2,                                

#if 1
    .period_bytes_min = 16,
    .period_bytes_max = __cap_buf_size,
    .periods_min = __cap_buf_cnt,
    .periods_max = __cap_buf_cnt,
    .buffer_bytes_max = __cap_buf_cnt * __cap_buf_size,
    .fifo_size = 16, //ignored
#else
    .period_bytes_min = 128,
    .period_bytes_max = __cap_buf_cnt * __cap_buf_size,
    .periods_min = 1,
    .periods_max = __cap_buf_cnt,
    .buffer_bytes_max = __cap_buf_cnt * __cap_buf_size,
    .fifo_size = 16, //ignored
#endif
};

/*** tea module ***/
static unsigned int io_ckc_get_dai_clock(unsigned int freq)
{
    switch (freq) {
    default:
    case 44100: return ((441 * 256) / 10);
    case 22000: return (((22050 * 256) / 100) / 10);
    case 11000: return (((11025 * 256) / 100) / 10);
    case 22050:
    case 11025:
    case 48000:
    case 24000:
    case 12000:
    case  8000:
    case 32000:
    case 16000: return (((freq * 256) / 100) / 10);
    }
    return 0;
}

static int alsa_get_intr_num(struct snd_pcm_substream *substream)
{
    if (substream) {
        return IRQ_ADMA;
    }
    return -1;
}

#if defined(CONFIG_SND_TCC_DAI2SPDIF)
void tea_i2s_setclock(unsigned int clock_rate, int is_enable_spdif)
{
    unsigned int ckc = 0;
    ckc = io_ckc_get_dai_clock(clock_rate);
    tca_ckc_setperi(PERI_DAI, 1, ckc * 10, DIRECTPLL2);
    if (is_enable_spdif) {
        tea_spdif_setclock(clock_rate);
    }
}
#else
void tea_i2s_setclock(unsigned int clock_rate)
{
    unsigned int ckc = 0;
    ckc = io_ckc_get_dai_clock(clock_rate);
    tca_ckc_setperi(PERI_DAI, 1, ckc * 10, DIRECTPLL2);
}
#endif
EXPORT_SYMBOL(tea_i2s_setclock);

void tea_spdif_setclock(unsigned int clock_rate)
{
    unsigned int daihz = 0;
    unsigned tmpCfg, tmpStatus;	
    volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);

    alsa_dbg("[%s], clock_rate[%u]\n", __func__, clock_rate);

    //daihz = ((clock_rate * 512) / 100);
    daihz = io_ckc_get_dai_clock(clock_rate);
    daihz = daihz * 2 * 10; /* set 512fs for HDMI */
    tca_ckc_setperi(PERI_SPDIF, ENABLE, daihz, DIRECTPLL2);

    tmpCfg = p_adma_spdif_tx_base->TxConfig;
    tmpStatus = p_adma_spdif_tx_base->TxChStat;

    if (clock_rate == 44100) { /* 44.1KHz */
        p_adma_spdif_tx_base->TxConfig = ((tmpCfg & 0xFFFF00FF) | (2 << 8));
        p_adma_spdif_tx_base->TxChStat = ((tmpStatus & 0xFFFFFF3F) | (0/*FREQ_44_1KHz*/ << 6));
    } else if (clock_rate == 48000) { /* 48KHz */
        p_adma_spdif_tx_base->TxConfig = ((tmpCfg & 0xFFFF00FF) | (2 << 8));
        p_adma_spdif_tx_base->TxChStat = ((tmpStatus & 0xFFFFFF3F) | (1/*FREQ_48KHz*/ << 6));
    } else if (clock_rate == 32000) { /* 32KHz */
        p_adma_spdif_tx_base->TxConfig = ((tmpCfg & 0xFFFF00FF) | (2 << 8));
        p_adma_spdif_tx_base->TxChStat = ((tmpStatus & 0xFFFFFF3F) | (2/*FREQ_32KHz*/ << 6));
    } else { /* Sampling Rate Converter */
        p_adma_spdif_tx_base->TxConfig = ((tmpCfg & 0xFFFF00FF) | (2 << 8));
        p_adma_spdif_tx_base->TxChStat = ((tmpStatus & 0xFFFFFF3F) | (3/*FREQ_SRC*/ << 6));
    }
}
EXPORT_SYMBOL(tea_spdif_setclock);

static void set_dma_outbuffer(unsigned int addr, unsigned int length, unsigned int period)
{
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    unsigned long dma_buffer = 0;

    alsa_dbg("[%s], addr[0x%08X], len[%u], period[%u]\n", __func__, addr, length, period);
    BITCLR(pADMA->ChCtrl, Hw28);

#if 0   /* Original source */
    dma_buffer = 0xFFFFFF00 / (((tcc_pcm_hardware_play.buffer_bytes_max) >> 4) << 8);
    dma_buffer = dma_buffer * (((tcc_pcm_hardware_play.buffer_bytes_max) >> 4) << 8);
#else
    dma_buffer = 0xFFFFFF00 / ((length >> 4) << 8);
    dma_buffer = dma_buffer * ((length >> 4) << 8);
#endif

    pADMA->TxDaParam = dma_buffer | 4;

    // generate interrupt when half of buffer was played
    pADMA->TxDaTCnt = (period >> 0x04) - 1;

    alsa_dbg("[%s] HwTxDaParam [0x%X]\n", __func__, (int)(dma_buffer | 4));
    alsa_dbg("[%s] HwTxDaTCnt [%d]\n", __func__, ((period) >> 0x04) - 1);
    pADMA->TxDaSar = addr;

    pADMA->TransCtrl |= (Hw28 | Hw16 | Hw9 | (Hw0 | Hw1) | Hw24);

    pADMA->RptCtrl = 0;
    pADMA->ChCtrl |= Hw12;
    //BITCLR(pADMA->RptCtrl, Hw31);
#if 0   /* Original source */
    if (length < tcc_pcm_hardware_play.buffer_bytes_max) {
        BITSET(pADMA->RptCtrl, (length / period) - 1);
    }
#endif

    //pADMA->ChCtrl |= (Hw28 | Hw12 | Hw0);
}

static void set_dma_spdif_outbuffer(unsigned int addr, unsigned int length, unsigned int period)
{
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);
    unsigned long dma_buffer = 0;

    alsa_dbg("%s, addr[0x%08X], len[%u], period[%u]\n", __func__, addr, length, period);
    BITCLR(pADMA->ChCtrl, Hw29);

    dma_buffer = 0xFFFFFF00 / (((length) >> 4) << 8);
    dma_buffer = dma_buffer * (((length) >> 4) << 8);

    pADMA->TxSpParam = dma_buffer | 4;

    // generate interrupt when half of buffer was played
    pADMA->TxSpTCnt = (period >> 0x04) - 1;

    alsa_dbg("[%s] HwTxDaParam [0x%X]\n", __func__, (int)(dma_buffer | 4));
    alsa_dbg("[%s] HwTxDaTCnt [%d]\n", __func__, ((period) >> 0x04) - 1);
    pADMA->TxSpSar = addr;

    pADMA->TransCtrl |= (Hw28 | Hw17 | Hw11 | (Hw3 | Hw2));

    pADMA->RptCtrl = 0;

    memset((void *)p_adma_spdif_tx_base->UserData, 0, 24);
    memset((void *)p_adma_spdif_tx_base->ChStatus, 0, 24);
    memset((void *)p_adma_spdif_tx_base->TxBuffer, 0, 128);

    p_adma_spdif_tx_base->TxConfig |= HwZERO | Hw8 | Hw1;
    p_adma_spdif_tx_base->TxBuffer[0] = 0;

    /*
     * Hw2: interrupt output enable
     * Hw1: data being valid
     * Hw0: enable the transmission
     */
    p_adma_spdif_tx_base->TxConfig |= HwZERO | Hw1 | Hw0;

    p_adma_spdif_tx_base->TxIntStat = 0x1E; /* Clear Interrupt Status */


    alsa_dbg("%s, spdif current addr[0x%08X] \n", __func__, pADMA->TxSpCsar);
    pADMA->ChCtrl |= (Hw13);
}

static void set_dma_inbuffer(unsigned int addr, unsigned int length, unsigned int period)
{
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    unsigned long dma_buffer = 0;

    alsa_dbg("[%s], addr[0x%08X], len[%d]\n", __func__, addr, length);

    BITCLR(pADMA->ChCtrl, Hw30);

#if 0   /* Original source */
    dma_buffer = 0xFFFFFF00 / (((tcc_pcm_hardware_capture.buffer_bytes_max) >> 4) << 8);
    dma_buffer = dma_buffer * (((tcc_pcm_hardware_capture.buffer_bytes_max) >> 4) << 8);
#else
    dma_buffer = 0xFFFFFF00 / ((length >> 4) << 8);
    dma_buffer = dma_buffer * ((length >> 4) << 8);
#endif

    //dma_buffer = 0xFFFFF8000;
    pADMA->RxDaParam = dma_buffer | 4;

    // generate interrupt when half of buffer was filled
    pADMA->RxDaTCnt = (period >> 0x04) - 1;

    alsa_dbg("[%s] HwRxDaParam [0x%X]\n", __func__, (int)dma_buffer | 4);
    alsa_dbg("[%s] HwRxDaTCnt [%d]\n", __func__, ((period) >> 0x04) - 1);
    pADMA->RxDaDar = addr;

    pADMA->TransCtrl |= (Hw29 | Hw18 | Hw13 | (Hw4 | Hw5) | Hw26);

    pADMA->RptCtrl = 0;

#if defined(CONFIG_SND_TCC_DAI2SPDIF)
    tcc_alsa_info.is_start_spdif = 0;
    set_dma_spdif_outbuffer(addr, length, period);
#endif

    pADMA->ChCtrl |= Hw14;
    //BITCLR(pADMA->RptCtrl, Hw31);
#if 0   /* Original source */
    if (length<tcc_pcm_hardware_capture.buffer_bytes_max) {
        BITSET(pADMA->RptCtrl, (length / period) - 1);
    }
#endif

    //pADMA->ChCtrl |= (Hw30 | Hw14 | Hw2);
}

#if defined(CONFIG_SND_TCC_DAI2SPDIF)
static int get_curdma_period_pos(unsigned int dma_addr, unsigned int cur_addr, unsigned int period_size, unsigned int total_size)
{
    if ((cur_addr >= dma_addr) && (cur_addr <= (dma_addr + total_size))) {
        unsigned int cur_size = cur_addr - dma_addr;

        if (cur_size) {
            return (cur_size / period_size);
        }
    }
    return 0;
}
#endif

static irqreturn_t tcc_dma_done_handler(int irq, void *dev_id)
{
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    struct snd_pcm_runtime *runtime;
    struct tcc_runtime_data *prtd;
#if 0
    dma_addr_t cur_ptr;
#endif
    int dmaInterruptSource = 0;

    unsigned long reg_temp = 0;                 

    if (pADMA->IntStatus & (Hw4 | Hw0)) {
        dmaInterruptSource |= DMA_CH_OUT;
        reg_temp |= Hw4 | Hw0;
    }
    if (pADMA->IntStatus & (Hw6 | Hw2)) {
        dmaInterruptSource |= DMA_CH_IN;
        reg_temp |= Hw6 | Hw2;
    }
    if (pADMA->IntStatus & (Hw5 | Hw1)) {
        dmaInterruptSource |= DMA_CH_SPDIF;
        reg_temp |= Hw5 | Hw1;
    }

    if (reg_temp) { 
        pADMA->IntStatus |= reg_temp;
    }

#if !defined(CONFIG_SND_TCC_DAI2SPDIF)
    if ((dmaInterruptSource & DMA_CH_SPDIF)
        && (tcc_alsa_info.flag & TCC_RUNNING_SPDIF)) {

        snd_pcm_period_elapsed(tcc_alsa_info.spdif_ptr);

        runtime = tcc_alsa_info.spdif_ptr->runtime;
        prtd = (struct tcc_runtime_data *)runtime->private_data;
    }
#endif

    if ((dmaInterruptSource & DMA_CH_OUT)
        && (tcc_alsa_info.flag & TCC_RUNNING_PLAY)) {
        snd_pcm_period_elapsed(tcc_alsa_info.play_ptr);

        runtime = tcc_alsa_info.play_ptr->runtime;
        prtd = (struct tcc_runtime_data *) runtime->private_data;

    #if 0
        cur_ptr = pADMA->TxDaCsar; 
        if (prtd->dma_buffer_end <= cur_ptr) {           
            BITSET(pADMA->ChCtrl, Hw28);
        }
    #endif
    }
    if ((dmaInterruptSource & DMA_CH_IN)
        && (tcc_alsa_info.flag & TCC_RUNNING_CAPTURE)) {
        snd_pcm_period_elapsed(tcc_alsa_info.cap_ptr);

#if defined(CONFIG_SND_TCC_DAI2SPDIF)
        if (tcc_alsa_info.is_start_spdif == 0) {
            runtime = tcc_alsa_info.cap_ptr->runtime;
            prtd = (struct tcc_runtime_data *)runtime->private_data;
            {
                int dai_pos = get_curdma_period_pos(prtd->dma_buffer,
                                                    pADMA->RxDaCdar,
                                                    prtd->period_size,
                                                    prtd->period_size * prtd->nperiod);

                int spdif_pos = get_curdma_period_pos(prtd->dma_buffer,
                                                      pADMA->TxSpCsar,
                                                      prtd->period_size,
                                                      prtd->period_size * prtd->nperiod);
                int is_start_condition = 0;

                if (dai_pos) {
                    if ((dai_pos - 1) == spdif_pos) { is_start_condition = 1; }
                } else {
                    if (spdif_pos == (prtd->nperiod - 1)) { is_start_condition = 1; }
                }

                if (is_start_condition) {
                    volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);

                    p_adma_spdif_tx_base->TxConfig |= HwZERO | Hw0 | Hw2 | Hw1;
                    BITSET(pADMA->ChCtrl, Hw1 | Hw29);

                    tcc_alsa_info.is_start_spdif = 1;
                    alsa_dbg("enable spdif  spdif_cur[0x%X], dai_cur[0x%X], dai_pos[%u], spdif_pos[%u]\n",
                               pADMA->TxSpCsar, pADMA->RxDaCdar, dai_pos, spdif_pos);
                } else {
                    alsa_dbg("skip spdif  spdif_cur[0x%X], dai_cur[0x%X], dai_pos[%u], spdif_pos[%u]!!!\n",
                               pADMA->TxSpCsar, pADMA->RxDaCdar, dai_pos, spdif_pos);
                }
            }
        }
#endif
        runtime = tcc_alsa_info.cap_ptr->runtime;
        prtd = (struct tcc_runtime_data *)runtime->private_data;

    #if 0
        cur_ptr = pADMA->RxDaCdar; 
        if (prtd->dma_buffer_end <= cur_ptr) {           
            BITSET(pADMA->ChCtrl, Hw30);
        }
    #endif
    }

    return IRQ_HANDLED;
}

static dma_addr_t get_dma_addr_offset(dma_addr_t start_addr, dma_addr_t cur_addr, unsigned int total_size, unsigned int dma_len)
{
    if ((start_addr <= cur_addr) && ((start_addr + total_size) > cur_addr)) {
        unsigned long remainder = cur_addr - start_addr;
        return (remainder - (remainder % dma_len));
    }
    return 0;
}

static int tcc_pcm_hw_params(struct snd_pcm_substream *substream, struct snd_pcm_hw_params *params)
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct tcc_runtime_data *prtd = runtime->private_data;
    struct snd_soc_pcm_runtime *rtd = substream->private_data;
    struct tcc_pcm_dma_params *dma = rtd->dai->cpu_dai->dma_data;
    size_t totsize = params_buffer_bytes(params);
    size_t period = params_period_bytes(params); 
    size_t period_bytes = params_period_bytes(params);
    int chs = params_channels(params);
    int ret = 0;
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    volatile PADMADAI pADMA_DAI = (volatile PADMADAI)tcc_p2v(HwADMA_DAIBASE);

    /* return if this is a bufferless transfer e.g.
     * codec <--> BT codec or GSM modem -- lg FIXME */
    if (!dma)
        return 0;      

    /* this may get called several times by oss emulation
     * with different params */
    if (prtd->params == NULL) {
        prtd->params = dma;

        mutex_lock(&(tcc_alsa_info.mutex));
        if (!(tcc_alsa_info.flag & TCC_INTERRUPT_REQUESTED)) {
            ret = request_irq(alsa_get_intr_num(substream),
                              tcc_dma_done_handler,
                              IRQF_SHARED,
                              "alsa-audio",
                              &tcc_alsa_info);
            if (ret < 0) {
                mutex_unlock(&(tcc_alsa_info.mutex));
                return -EBUSY;
            }
            tcc_alsa_info.flag |= TCC_INTERRUPT_REQUESTED;
        }
        mutex_unlock(&(tcc_alsa_info.mutex));

        if (substream->pcm->device == __SPDIF_DEV_NUM__) {
            prtd->dma_ch = DMA_CH_SPDIF;
        } else {
            prtd->dma_ch = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ? DMA_CH_OUT : DMA_CH_IN;
        }
    } else {
        /* should works more? */
    }        

    //DMA  Ʈ
    //DMA    ʱȭ 

    snd_pcm_set_runtime_buffer(substream, &substream->dma_buffer);        
    runtime->dma_bytes = totsize;

    prtd->dma_buffer_end = runtime->dma_addr + totsize;
    prtd->dma_buffer = runtime->dma_addr;
    prtd->period_size = period_bytes;
    prtd->nperiod = period;

    alsa_dbg("\n totsize=0x%X period=0x%X  period num=%d, rate=%d\n",
             totsize, period_bytes, period, params_rate(params));

    {
            dma_addr_t offset = 0;
            dma_addr_t cur_addr = 0;
            unsigned int total_dma_size = 0;

            if (substream->pcm->device == __SPDIF_DEV_NUM__) {
                cur_addr = pADMA->TxSpCsar;
                total_dma_size = tcc_pcm_hardware_play.buffer_bytes_max;
            } else {
                cur_addr = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ? pADMA->TxDaCsar : pADMA->RxDaCdar;
                total_dma_size = tcc_pcm_hardware_capture.buffer_bytes_max;
            }

            offset = get_dma_addr_offset(prtd->dma_buffer, cur_addr, total_dma_size, totsize);

            prtd->dma_buffer += offset;
            prtd->dma_buffer_end = prtd->dma_buffer + totsize;

            runtime->dma_addr = prtd->dma_buffer;
            runtime->dma_area += offset;

            alsa_dbg("########## dma_addr[0x%X], cur_addr[0x%X], totsize[0x%X], period[0x%X]\n",
                     runtime->dma_addr, cur_addr, totsize, period);
    }

    if (substream->pcm->device == __SPDIF_DEV_NUM__) {
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
            set_dma_spdif_outbuffer(prtd->dma_buffer, totsize, period_bytes);
        } else {
            printk("cannot support S/PDIF capture !!!!!\n");
            return -1;
        }
    } else {
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
            set_dma_outbuffer(prtd->dma_buffer, totsize, period_bytes);
        } else {
            set_dma_inbuffer(prtd->dma_buffer, totsize, period_bytes);
        }
    }

    if (chs > 2) {
        pADMA_DAI->DAMR |= (Hw29 | Hw28);
        pADMA->RptCtrl |= (Hw26 | Hw25 | Hw24);
        pADMA->ChCtrl |= (Hw24 | Hw21 | Hw20);
    } else {
        BITCLR(pADMA_DAI->DAMR, (Hw29 | Hw28));
        BITCLR(pADMA->RptCtrl, (Hw26 | Hw25 | Hw24));
        BITCLR(pADMA->ChCtrl,(Hw24 | Hw21 | Hw20));
    }
    return ret;
}

static int tcc_pcm_hw_free(struct snd_pcm_substream *substream)
{
    struct tcc_runtime_data *prtd = substream->runtime->private_data;

    if (prtd->dma_ch) {
        snd_pcm_set_runtime_buffer(substream, NULL);
        //߰ ʿ  ??
    }

    return 0;
}

static int tcc_pcm_prepare(struct snd_pcm_substream *substream)
{
    struct snd_pcm_runtime *runtime = substream->runtime;

    //DMA initialize

    /* Disable Noise */
    //memset((void *)(&HwDADO_L0), 0, 32);
    memset(runtime->dma_area, 0x00, runtime->dma_bytes);

    return 0;
}

static int tcc_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
    //struct tcc_runtime_data *prtd = substream->runtime->private_data;
    //    audio_stream_t *s = &output_stream;        
    int ret = 0;
    volatile PADMADAI pDAI = (volatile PADMADAI)tcc_p2v(HwADMA_DAIBASE);
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
	volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);

    switch (cmd) {
    case SNDRV_PCM_TRIGGER_START:
    case SNDRV_PCM_TRIGGER_RESUME:        
    case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
        if (substream->pcm->device == __SPDIF_DEV_NUM__) {
#if 0
            //tca_i2s_start(pDAI, p_adma_spdif_tx_base, 0);
            p_adma_spdif_tx_base->TxIntStat = 0x1E;		/* Clear Interrupt Status */
            p_adma_spdif_tx_base->TxConfig |= HwZERO | Hw0;
            BITSET(pADMA->ChCtrl, Hw1 | Hw29);
#else
            p_adma_spdif_tx_base->TxConfig |= HwZERO | Hw0 | Hw2 | Hw1;
            BITSET(pADMA->ChCtrl, Hw1 | Hw29);
#endif
        } else {
            unsigned long tmp = 0;
            if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
                //tca_i2s_start(pDAI, p_adma_spdif_tx_base, 0);
                BITSET(pADMA->ChCtrl, Hw0 | Hw28);
                tmp = pDAI->DAVC;
                pDAI->DAVC = 0x10;
                pDAI->DAMR |= Hw14;
                mdelay(10);
                pDAI->DAVC = tmp;

            } else {
                //tca_i2s_start(pDAI, p_adma_spdif_tx_base, 1);
                BITSET(pADMA->ChCtrl, Hw2 | Hw30);
                tmp = pDAI->DAVC;
                pDAI->DAVC = 0x10;
                pDAI->DAMR |= Hw13;
#if defined(CONFIG_SND_TCC_DAI2SPDIF)
                tcc_alsa_info.is_start_spdif = 0;
#endif
                mdelay(10);
                pDAI->DAVC = tmp;
            }
        }
        break;

    case SNDRV_PCM_TRIGGER_STOP:
    case SNDRV_PCM_TRIGGER_SUSPEND:
    case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
        if (substream->pcm->device == __SPDIF_DEV_NUM__) {
#if 0
            //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 0);
            p_adma_spdif_tx_base->TxConfig &= ~Hw0;
            BITCLR(pADMA->ChCtrl, Hw1);
#else
            p_adma_spdif_tx_base->TxConfig &= ~(Hw0 | Hw2 | Hw1);
            BITCLR(pADMA->ChCtrl, Hw1 | Hw29);
#endif
        } else {
            if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
                //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 0);
                pDAI->DAMR &= ~Hw14;
                BITCLR(pADMA->ChCtrl, Hw0 | Hw28);
            } else {
                //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 1);
                pDAI->DAMR &= ~Hw13;
                BITCLR(pADMA->ChCtrl, Hw2 | Hw30);
#if defined(CONFIG_SND_TCC_DAI2SPDIF)
                p_adma_spdif_tx_base->TxConfig &= ~(Hw0 | Hw2 | Hw1);
                BITCLR(pADMA->ChCtrl, Hw1 | Hw29);
                tcc_alsa_info.is_start_spdif = 0;
#endif
            }
        }
        break;

    default:
        ret = -EINVAL;
    }

    return ret;
}


//buffer point update
//current position   range 0-(buffer_size-1)
static snd_pcm_uframes_t tcc_pcm_pointer(struct snd_pcm_substream *substream)
{
    volatile PADMA pADMA = (volatile PADMA)tcc_p2v(HwADMA_BASE);
    struct snd_pcm_runtime *runtime = substream->runtime;
    //struct tcc_runtime_data *prtd = runtime->private_data;
    snd_pcm_uframes_t x;
    dma_addr_t ptr = 0;

    if (substream->pcm->device == __SPDIF_DEV_NUM__) {
        ptr = pADMA->TxSpCsar;
    } else {
        ptr = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ? pADMA->TxDaCsar : pADMA->RxDaCdar;
    }

    x = bytes_to_frames(runtime, ptr - runtime->dma_addr);

    if (x < runtime->buffer_size) {
        return x;
    }
    return 0;
}

static int tcc_pcm_open(struct snd_pcm_substream *substream)
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct tcc_runtime_data *prtd;
    int ret;

    mutex_lock(&(tcc_alsa_info.mutex));
    if (substream->pcm->device == __SPDIF_DEV_NUM__) {
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
            tcc_alsa_info.flag |= TCC_RUNNING_SPDIF;
            tcc_alsa_info.spdif_ptr = substream;
            snd_soc_set_runtime_hwparams(substream, &tcc_pcm_hardware_play);
            alsa_dbg("[%s] open spdif device\n", __func__);
        } else {
            printk("cannot support S/PDIF capture !!!!!\n");
            return -1;
        }
    } else {
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {  
            tcc_alsa_info.flag |= TCC_RUNNING_PLAY;
            tcc_alsa_info.play_ptr = substream;
            snd_soc_set_runtime_hwparams(substream, &tcc_pcm_hardware_play);
        } else {
            tcc_alsa_info.flag |= TCC_RUNNING_CAPTURE;
            tcc_alsa_info.cap_ptr = substream;
            snd_soc_set_runtime_hwparams(substream, &tcc_pcm_hardware_capture);
        }
    }
    mutex_unlock(&(tcc_alsa_info.mutex));

    prtd = kzalloc(sizeof(struct tcc_runtime_data), GFP_KERNEL);

    if (prtd == NULL) {
        ret = -ENOMEM;
        goto out;
    }
    runtime->private_data = prtd;

    return 0;
    kfree(prtd);
out:
    return ret;
}

static int tcc_pcm_close(struct snd_pcm_substream *substream)
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct tcc_runtime_data *prtd = runtime->private_data;
    volatile PADMADAI pDAI = (volatile PADMADAI)tcc_p2v(HwADMA_DAIBASE);
    volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);

    mutex_lock(&(tcc_alsa_info.mutex));
    if (substream->pcm->device == __SPDIF_DEV_NUM__) {
        tcc_alsa_info.flag &= ~TCC_RUNNING_SPDIF;
        //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 0);
        p_adma_spdif_tx_base->TxConfig &= ~Hw0;
        alsa_dbg("[%s] close spdif device\n", __func__);
    } else {
        if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
            tcc_alsa_info.flag &= ~TCC_RUNNING_PLAY;
            //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 0);
            pDAI->DAMR &= ~Hw14;
        } else {
            tcc_alsa_info.flag &= ~TCC_RUNNING_CAPTURE;
            //tca_i2s_stop(pDAI, p_adma_spdif_tx_base, 1);
            pDAI->DAMR &= ~Hw13;
#if defined(CONFIG_SND_TCC_DAI2SPDIF)
            p_adma_spdif_tx_base->TxConfig &= ~Hw0;
            tcc_alsa_info.is_start_spdif = 0;
#endif
        }
    }
    // dma_free_writecombine(substream->pcm->card->dev, PAGE_SIZE,); 

    if (prtd) {
        kfree(prtd);
    }

    if (tcc_alsa_info.flag & TCC_INTERRUPT_REQUESTED) {
        if (!(tcc_alsa_info.flag & (TCC_RUNNING_SPDIF | TCC_RUNNING_PLAY | TCC_RUNNING_CAPTURE))) {
            free_irq(alsa_get_intr_num(substream), &tcc_alsa_info);
            tcc_alsa_info.flag &= ~TCC_INTERRUPT_REQUESTED;
        }
    }
    mutex_unlock(&(tcc_alsa_info.mutex));

    return 0;
}

static int tcc_pcm_mmap(struct snd_pcm_substream *substream, struct vm_area_struct *vma)
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    return dma_mmap_writecombine(substream->pcm->card->dev, vma,
        runtime->dma_area,
        runtime->dma_addr,
        runtime->dma_bytes);
}

struct snd_pcm_ops tcc_pcm_ops = {
    .open  = tcc_pcm_open,
    .close  = tcc_pcm_close,
    .ioctl  = snd_pcm_lib_ioctl,
    .hw_params = tcc_pcm_hw_params,
    .hw_free = tcc_pcm_hw_free,
    .prepare = tcc_pcm_prepare,
    .trigger = tcc_pcm_trigger,
    .pointer = tcc_pcm_pointer,
    .mmap = tcc_pcm_mmap,
};

static int tcc_pcm_preallocate_dma_buffer(struct snd_pcm *pcm, int stream)
{
    struct snd_pcm_substream *substream = pcm->streams[stream].substream;
    struct snd_dma_buffer *buf = &substream->dma_buffer;
    size_t size = 0;

    if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
        size = tcc_pcm_hardware_play.buffer_bytes_max;
    } else {
        size = tcc_pcm_hardware_capture.buffer_bytes_max;
    }

    buf->dev.type = SNDRV_DMA_TYPE_DEV;
    buf->dev.dev = pcm->card->dev;
    buf->private_data = NULL;

    alsa_dbg("tcc_pcm_preallocate_dma_buffer size [%d]\n", size);
    buf->area = dma_alloc_writecombine(pcm->card->dev, size, &buf->addr, GFP_KERNEL);
    if (!buf->area) {
        return -ENOMEM;
    }

    buf->bytes = size;

    return 0;
}


static void tcc_pcm_free_dma_buffers(struct snd_pcm *pcm)
{
    struct snd_pcm_substream *substream;
    struct snd_dma_buffer *buf;
    int stream;

    for (stream = 0; stream < 2; stream++) {
        substream = pcm->streams[stream].substream;
        if (!substream) { continue; }

        buf = &substream->dma_buffer;
        if (!buf->area) { continue; }

        dma_free_writecombine(pcm->card->dev, buf->bytes, buf->area, buf->addr);
        buf->area = NULL;
    }
    mutex_init(&(tcc_alsa_info.mutex));
}

static u64 tcc_pcm_dmamask = DMA_32BIT_MASK;

int tcc_pcm_new(struct snd_card *card, struct snd_soc_dai *dai, struct snd_pcm *pcm)
{
    int ret = 0;

    memset(&tcc_alsa_info, 0, sizeof(tcc_interrupt_info_x));
    mutex_init(&(tcc_alsa_info.mutex));

    if (!card->dev->dma_mask) {
        card->dev->dma_mask = &tcc_pcm_dmamask;
    }
    if (!card->dev->coherent_dma_mask) {
        card->dev->coherent_dma_mask = DMA_32BIT_MASK;
    }
    if (dai->playback.channels_min) {
        ret = tcc_pcm_preallocate_dma_buffer(pcm,SNDRV_PCM_STREAM_PLAYBACK);
        if (ret) { goto out; }
    }

    if (dai->capture.channels_min) {
        ret = tcc_pcm_preallocate_dma_buffer(pcm,SNDRV_PCM_STREAM_CAPTURE);
        if (ret) { goto out; }
    } 
	{
        volatile ADMASPDIFTX *p_adma_spdif_tx_base = (volatile ADMASPDIFTX *)tcc_p2v(HwADMA_SPDIFTXBASE);

        memset((void *)p_adma_spdif_tx_base->UserData, 0, 24);
        memset((void *)p_adma_spdif_tx_base->ChStatus, 0, 24);
        memset((void *)p_adma_spdif_tx_base->TxBuffer, 0, 128);

        //unsigned int bitSet = 0x0;

        p_adma_spdif_tx_base->TxConfig |= HwZERO
            //16bits 
            |Hw8 //Clock Divider Ratio is 2(384*fs to default)
            //| Hw7 | Hw6 //User Data Enalbe Bits is set reserved
            //| Hw5 | Hw4 //Channel Status Enalbe Bits is set reserved
            //| Hw2 //interrupt output enable
            | Hw1 //data being valid
            //| Hw0 //enable the transmission
            ;
        //Set SPDIF Transmit Channel Status
        p_adma_spdif_tx_base->TxChStat |= HwZERO
            //| 0x00000400 // Store Channel Status bit
            //| 0x00000200 // Store User Data bit
            //| 0x00000100 // Store Validity bit
            //44.1kHz
            //| Hw3 //Original/Commercially Pre-recored data
            //| Hw2 //Pre-emphasis is 50/15us
            //| Hw1 //Copy permitted
            //Audio format
            ;		

        //p_adma_spdif_tx_base->TxIntMask = HwZERO
        //	| Hw2//Higher Data buffer is empty
        //	| Hw1//Lower Data buffer is empty
        //	;

        p_adma_spdif_tx_base->DMACFG = Hw7;
        msleep(100);

        p_adma_spdif_tx_base->DMACFG = HwZERO
            //| Hw14	//Swap Sample Enable
            | Hw13 	//Read Address 16bit Mode
            | Hw11	//DMA Request enable for user data buffer
            | Hw10	//DMA Request enable for sample data buffer
            //| Hw8	//Enable address
            //| Hw3	//FIFO Threshold for DMA request
            | Hw1| Hw0	// [3:0] FIFO Threshold for DMA Request
            ;

        /* Initialize Sample Data Buffer */
        //while(p_adma_spdif_tx_base->DMACFG & 0x00300000) p_adma_spdif_tx_base->TxBuffer[0] = 0;
        p_adma_spdif_tx_base->TxBuffer[0] = 0;

        p_adma_spdif_tx_base->TxConfig |= HwZERO
            //| Hw2 //interrupt output enable
            | Hw1 //data being valid
            | Hw0 //enable the transmission
            ;

#if 0
        while((p_adma_spdif_tx_base->DMACFG & 0x00100000)); /* Wait until first data TX */
        while((p_adma_spdif_tx_base->DMACFG & 0x001F0000) > 0x00080000);

        while(!(p_adma_spdif_tx_base->TxIntStat & Hw1));	/* Wait for Interrupt Status */
#endif

        p_adma_spdif_tx_base->TxIntStat = 0x1E; /* Clear Interrupt Status */

#if 0
        p_adma_spdif_tx_base->TxConfig |= HwZERO
            | Hw2 //interrupt output enable
            | Hw1 //data being valid
            //| Hw0 //enable the transmission
            ;

        // Tx channel Control Register Setting	
        //Start DMA
        bitSet = ( HwZERO
                   |Hw1 //Interrupt Enable of SPDIF Tx	
                 );

        pADMA->ChCtrl |= bitSet;
#endif
    }


out:
    return ret;
}

#if 0
int tcc_pcm_suspend(struct platform_device *pdev, struct snd_soc_cpu_dai *cpu_dai)
{
    return 0;
}

int tcc_pcm_resume(struct platform_device *pdev, struct snd_soc_cpu_dai *cpu_dai)
{
    return 0;
}
#endif

struct snd_soc_platform tcc_soc_platform = {
    .name  = "tcc-audio",
    .pcm_ops  = &tcc_pcm_ops,
    .pcm_new = tcc_pcm_new,
    .pcm_free = tcc_pcm_free_dma_buffers,
#if 0
    .suspend = tcc_pcm_suspend,
    .resume = tcc_pcm_resume,
#endif
};

EXPORT_SYMBOL_GPL(tcc_soc_platform);


#if !defined(CONFIG_SND_TCC_DAI2SPDIF)
/************************************************************************
 * Dummy function for S/PDIF. This is a codec part.
************************************************************************/
static int iec958_pcm_prepare(struct snd_pcm_substream *substream) { return 0; }
static int iec958_hw_params(struct snd_pcm_substream *substream, struct snd_pcm_hw_params *params) { return 0; }
static void iec958_shutdown(struct snd_pcm_substream *substream) { }
static int iec958_mute(struct snd_soc_dai *dai, int mute) { return 0; }
static int iec958_set_dai_sysclk(struct snd_soc_dai *codec_dai, int clk_id, unsigned int freq, int dir) { return 0; }
static int iec958_set_dai_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt) { return 0; }

struct snd_soc_dai iec958_dai = {
	.name = "IEC958",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 1,
		.channels_max = 2,
        .rates = (SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 | SNDRV_PCM_RATE_48000),
		.formats = (SNDRV_PCM_FMTBIT_S16_LE),},
	.ops = {
		.prepare = iec958_pcm_prepare,
		.hw_params = iec958_hw_params,
		.shutdown = iec958_shutdown,
		.digital_mute = iec958_mute,
		.set_sysclk = iec958_set_dai_sysclk,
		.set_fmt = iec958_set_dai_fmt,
	}
};
EXPORT_SYMBOL_GPL(iec958_dai);

static int __init iec958_modinit(void)
{
	return snd_soc_register_dai(&iec958_dai);
}
module_init(iec958_modinit);

static void __exit iec958_exit(void)
{
	snd_soc_unregister_dai(&iec958_dai);
}
module_exit(iec958_exit);
/***********************************************************************/
#endif


/* S: Kernel v2.6.29, sgbaek */
static int __init tcc_soc_platform_init(void)
{
	return snd_soc_register_platform(&tcc_soc_platform);
}
module_init(tcc_soc_platform_init);

static void __exit tcc_soc_platform_exit(void)
{
	snd_soc_unregister_platform(&tcc_soc_platform);
}
module_exit(tcc_soc_platform_exit);
/* E: Kernel v2.6.29, sgbaek */


MODULE_AUTHOR("Telechips");
MODULE_DESCRIPTION("Telechips TCC PCM DMA module");
MODULE_LICENSE("GPL");

